
import requests
from distutils.version import StrictVersion
from distutils.version import LooseVersion
import sys
import os
import subprocess
from . import __version__
from . import windows
from .qt_task import Task

from PyQt5.QtWidgets import QMessageBox


# from https://gist.github.com/trinitronx/026574839d96a0c0efe1e9f2fd300f03
def get_versions(PKG_NAME='spectramanipulator'):
    """Get all versions of a package from PyPI"""
    r = requests.get(f'https://pypi.org/pypi/{PKG_NAME}/json')
    data = r.json()
    versions = data['releases'].keys()
    # remove dev versions
    versions = list(filter(lambda v: 'rc' not in v and 'dev' not in v and 'post' not in v, versions))
    try:
        versions.sort(key=StrictVersion)
    except ValueError as e:
        if 'invalid version number' in str(e):
            versions.sort(key=LooseVersion)
        else:
            raise e

    return versions


def get_latest_version():
    return get_versions()[-1]

# https://stackoverflow.com/questions/11887762/how-do-i-compare-version-numbers-in-python
def is_latest_version(ver):
    return LooseVersion(__version__) >= LooseVersion(ver)
    # return LooseVersion("0.1.0") >= LooseVersion(ver)


# # https://stackoverflow.com/questions/636561/how-can-i-run-an-external-command-asynchronously-from-python
class TaskUpdate(Task):
    def __init__(self, parent=None):
        super(TaskUpdate, self).__init__(parent)
        self.parent = parent
        self.latest_version = __version__

    def can_update(self):
        try:
            self.latest_version = get_latest_version()
        except requests.exceptions.RequestException as e:
            QMessageBox.critical(self.parent, 'Update', f"Connection error.\n{str(e)}")
            return False

        if is_latest_version(self.latest_version):
            QMessageBox.information(self.parent, 'Update', "The program is already up to date.")
            return False

        reply = QMessageBox.question(self.parent, 'Update',
                                     f"New version {self.latest_version} was found. Current version is {__version__}. "
                                     + "Do you want to update the program?", QMessageBox.Yes |
                                     QMessageBox.No)

        if reply == QMessageBox.No:
            return False

        if sys.platform == 'win32':
            windows.set_attached_console_visible(True)

        return True

    def run(self):
        python_dir = os.path.dirname(sys.executable)
        pip_path = os.path.join(python_dir, "Scripts", "pip.exe")

        subprocess.run(f"{pip_path} install spectramanipulator=={self.latest_version}")

    def postRun(self):
        if sys.platform == 'win32':
            windows.set_attached_console_visible(False)
        QMessageBox.information(self.parent, 'Update',
                                "The program was updated. Please restart your application (all instances have to be closed).")




import re


def rename(expression, text, counter=0):
    """
    TODO -->>> rewrite
    Returns the expression text specially parsed. It can contain two types of expression,
    counter term and slice term. The counter term can only be in python number format,
    defined here https://pyformat.info/. Eg. {:4g} - counter will be formatted and this term
    will be replaced by the number formatted in this way, eg. if counter = 25, {:4g} will
    be replaced by 0025.

    Other type is slice expression {start_index:end_index}. This term will be replaced by
    the text[start_index:end_index] value. It is the python slicing format for lists/strings/arrays.

    If {} brackets need to be in the resulting string, backslash must be used before the bracket,
    indicating it is not the term.

    ---------Example------

    expression = r'\{{:03d}\}: t = {1:} us'

    text = #164.6

    counter = 39

    result = r'{039}: t = 164.6 us'

    :param expression:
        str
    :param text:
        str
    :param counter:
        int
    :return:
        str
    """

    # copied from https://stackoverflow.com/questions/680826/python-create-slice-object-from-string
    def parse_slice(v):
        """
        Parses text like python "slice" expression (ie ``-10::2``).

        :param v:
            the slice expression or a lone integer
        :return:
            - None if input is None/empty
            - a ``slice()`` instance (even if input a lone numbrt)
        :raise ValueError:
            input non-empty but invalid syntax
        """
        orig_v = v
        v = v and v.strip()
        if not v:
            return

        try:
            if ':' not in v:
                ## A lone number given.
                v = int(v)
                return slice(v, v + 1)

            return slice(*map(lambda x: int(x.strip()) if x.strip() else None,
                              v.split(':')))
        except:
            raise ValueError("Parsing error of slice {}.".format(orig_v))

        # ## An alternative is to return `slice(None)` here.
        # raise trt.TraitError("Syntax-error in '%s' slice!" % orig_v)

    try:
        start_idx = 0
        contents = []

        for i, char in enumerate(expression):
            if char == '{' and i != 0 and expression[i - 1] != '\\':
                start_idx = i

            if char == '}' and expression[i - 1] != '\\':
                contents.append([start_idx, i + 1, expression[start_idx + 1:i]])

        if len(contents) == 0:
            return expression.replace('\\{', '{').replace('\\}', '}')

        # print the beginning
        result = expression[:contents[0][0]]

        for i, con in enumerate(contents):

            # for different formats for numbers, look https://pyformat.info/
            if 'd' in con[2] or 'f' in con[2] or 'g' in con[2]:
                content = ("{" + con[2] + "}").format(int(counter) if 'd' in con[2] else counter)
            else:  # slicing text
                content = text[parse_slice(con[2])]

            result += content + (expression[con[1]:contents[i + 1][0]] if i < len(contents) - 1 else '')

        # print the last part
        result += expression[contents[-1][1]:]

        return result.replace('\\{', '{').replace('\\}', '}')
    except:
        raise ValueError("Parsing error of expression {}.".format(expression))


#
# names = ['4865', '44555', "#4686486", "4adads"]
#
# # expression = "\{{:03d}\}: t = {1:} us"
# expression = "{1:5}"
#
# pattern = re.compile(r'{(\d+)?:(\d+)?:?(\d+)?}')
# m = pattern.search(expression)
#
# for i in range(len(names)):
#
#     print(names[i], "     formatted to      ", sl_text[sl])
#
#



#









from PyQt5.QtCore import Qt, QItemSelectionModel, QItemSelection, pyqtSignal, QModelIndex
from PyQt5.QtWidgets import QApplication, QMessageBox, QMenu, QAction
from PyQt5.QtGui import QCursor, QColor
import numpy as np
import os


from spectramanipulator.spectrum import Spectrum, SpectrumList

from spectramanipulator.dialogs.int_int_inputdialog import IntIntInputDialog
from spectramanipulator.dialogs.interpolate_dialog import InterpolateDialog
from spectramanipulator.dialogs.rename_dialog import RenameDialog
from spectramanipulator.dialogs.fitwidget import FitWidget
from spectramanipulator.dialogs.stylewidget import StyleWidget
# from dialogs.rangedialog import RangeDialog
from spectramanipulator.dialogs.rangewidget import RangeWidget
from spectramanipulator.dialogs.export_spectra_as import ExportSpectraAsDialog

from spectramanipulator.settings import Settings
from spectramanipulator.logger import Logger
from spectramanipulator.utils.rename import rename

from spectramanipulator.treeview.item import SpectrumItemGroup, SpectrumItem, GenericItem
from spectramanipulator.treeview.model import TreeView, ItemIterator

from spectramanipulator.console import Console

from spectramanipulator.parsers import parse_XML_Spreadsheet
from spectramanipulator.dataloader import parse_text, parse_files
from spectramanipulator.exporter import list_to_string, list_to_files
# from spectramanipulator.plotwidget import PlotWidget


# TODO-->> rewrite
def get_hierarchic_list(items_iterator):
    """ get a hierarchic structure of selected items
    spectra in groups are appended in list of spectrum objects
    spectra not in groups are appended as spectrum objects
    """

    sp_list = []
    temp_list = []
    last_grp_item = None

    for item in items_iterator:
        if isinstance(item, SpectrumItemGroup):
            continue

        item.name = item.name
        if item.is_in_group():
            curr_grp_item = item.parent
            item.group_name = curr_grp_item.name
            if last_grp_item != curr_grp_item:
                last_grp_item = curr_grp_item
                if len(temp_list) != 0:
                    sp_list.append(temp_list)
                    temp_list = []
                temp_list.append(item)
            else:
                temp_list.append(item)
        else:
            if len(temp_list) > 0:
                sp_list.append(temp_list)
                temp_list = []
            item.group_name = None
            sp_list.append(item)

    if len(temp_list) > 0:
        sp_list.append(temp_list)

    return sp_list


class TreeWidget(TreeView):
    redraw_spectra = pyqtSignal()
    state_changed = pyqtSignal()

    # all_spectra_list =

    def __init__(self, parent=None):
        super(TreeWidget, self).__init__(parent)

        self.main_widget = self.parent().parent()

        self.header().setStretchLastSection(True)

        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self.context_menu)

        self.items_deleted_signal.connect(self.items_deleted)
        self.myModel.item_edited_signal.connect(self.item_edited)
        self.myModel.checked_changed_signal.connect(self.check_changed)
        self.myModel.data_dropped_signal.connect(self.data_dropped)
        self.myModel.all_unchecked_signal.connect(lambda: self.redraw_spectra.emit())
        self.myModel.data_modified_signal.connect(self.data_modified)
        self.myModel.info_modified_signal.connect(self.info_modified)
        self.myModel.items_ungrouped_signal.connect(lambda: self.redraw_spectra.emit())

    def items_deleted(self, item_was_checked):
        if item_was_checked:
            self.redraw_spectra.emit()

    def item_edited(self, item_is_checked):
        self.state_changed.emit()
        if item_is_checked:
            self.redraw_spectra.emit()

    def data_modified(self, items):
        self.main_widget.update_items_data(items)

    def info_modified(self, items):
        self.setup_info()
        self.update_view()

    def check_changed(self, items, checked):
        # self.redraw_spectra.emit()
        self.main_widget.redraw_items(items, not checked)

    def data_dropped(self):
        self.redraw_spectra.emit()

    def save_state(self):
        super(TreeWidget, self).save_state()
        self.state_changed.emit()

        # if self.top_level_items_count() == 0:
        #     self.all_spectra_list = []
        #     return

        # self.all_spectra_list = self.get_hierarchic_list(
        #     self.myModel.iterate_items(ItemIterator.NoChildren))

        # Console.push_variables({'item': self.all_spectra_list})
        # Console.push_variables({'item': self.myModel.root})

    def export_selected_items_as(self):

        if ExportSpectraAsDialog.is_opened:
            ExportSpectraAsDialog.get_instance().activateWindow()
            ExportSpectraAsDialog.get_instance().setFocus()
            return

        if len(self.selectedIndexes()) == 0:
            return

        dialog = ExportSpectraAsDialog()

        if not dialog.accepted:
            return

        path, ext, delimiter, decimal_sep = dialog.result

        sp_list = get_hierarchic_list(
            self.myModel.iterate_selected_items(skip_groups=True,
                                                skip_childs_in_selected_groups=False))

        try:
            list_to_files(sp_list, path, ext,
                          include_group_name=Settings.files_exp_include_group_name,
                          include_header=Settings.files_exp_include_header,
                          delimiter=delimiter,
                          decimal_sep=decimal_sep,
                          x_data_name=Settings.bottom_axis_label)

        except Exception as ex:
            QMessageBox.warning(self, 'Error', ex.__str__(), QMessageBox.Ok)

        Logger.message(f"Data were saved to {path}")

    def copy_selected_items_to_clipboard(self):

        sp_list = get_hierarchic_list(self.myModel.iterate_selected_items(skip_groups=True,
                                                                          skip_childs_in_selected_groups=False))

        if len(sp_list) == 0:
            return

        Logger.status_message("Copying selected items to clipboard...")
        try:

            output = list_to_string(sp_list, include_group_name=Settings.clip_exp_include_group_name,
                                    include_header=Settings.clip_exp_include_header,
                                    delimiter=Settings.clip_exp_delimiter,
                                    decimal_sep=Settings.clip_exp_decimal_sep,
                                    x_data_name=Settings.bottom_axis_label)
            cb = QApplication.clipboard()
            cb.clear(mode=cb.Clipboard)
            cb.setText(output, mode=cb.Clipboard)

        except Exception as ex:
            Logger.message(ex.__str__())
            return

        Logger.status_message("Done")

    def paste_from_clipboard(self):

        m = QApplication.clipboard().mimeData()

        if m is not None and m.hasFormat("XML Spreadsheet") and not Settings.excel_imp_as_text:
            self.parse_XML_Spreadsheet(m.data("XML Spreadsheet").data())
            return

        cb = QApplication.clipboard()
        text_data = cb.text(mode=cb.Clipboard)
        spectra = parse_text(text_data)

        self.import_spectra(spectra)

    # --------- Operation Functions --------

    def normalize(self):

        if len(self.selectedIndexes()) == 0:
            return

        def accepted():
            x0, x1 = rng_dialog.returned_range

            try:
                items = []
                for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                                skip_childs_in_selected_groups=False):
                    item.normalize_no_update(x0, x1)
                    items.append(item)

                self.data_modified(items)
                self.state_changed.emit()

            except ValueError as ex:
                Logger.message(ex.__str__())
                QMessageBox.warning(self, 'Warning', ex.__str__(), QMessageBox.Ok)
                return

            Logger.status_message("Done")

        rng_dialog = RangeWidget(self.main_widget.var_widget, accepted, title="Normalize",
                                 label_text="Set x range values. Maximum y value is find in this range and y values "
                                            "are divided by this maximum. For normalizing to specific value, set x0 "
                                            "equal to x1:",
                                 parent=self)

    def cut(self):

        if len(self.selectedIndexes()) == 0:
            return

        def accepted():
            x0, x1 = rng_dialog.returned_range

            # Logger.status_message("Cutting the spectra...")

            try:
                items = []
                for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                                skip_childs_in_selected_groups=False):
                    item.cut_no_update(x0, x1)
                    items.append(item)

                self.data_modified(items)
                self.info_modified(items)
                self.state_changed.emit()

            except ValueError as ex:
                Logger.message(ex.__str__())
                QMessageBox.warning(self, 'Warning', ex.__str__(), QMessageBox.Ok)
                return

            Logger.status_message("Done")

        rng_dialog = RangeWidget(self.main_widget.var_widget, accepted, title="Cut",
                                 label_text="Set x range values. Values outside this range will"
                                            " be deleted from the spectrum/spectra:",
                                 parent=self)

    def extend_by_zeros(self):

        if len(self.selectedIndexes()) == 0:
            return

        def accepted():
            x0, x1 = rng_dialog.returned_range

            try:
                items = []
                for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                                skip_childs_in_selected_groups=False):
                    item.extend_by_value_no_update(x0, x1)
                    items.append(item)

                self.data_modified(items)
                self.info_modified(items)
                self.state_changed.emit()

            except ValueError as ex:
                Logger.message(ex.__str__())
                QMessageBox.warning(self, 'Warning', ex.__str__(), QMessageBox.Ok)
                return

            Logger.status_message("Done")

        rng_dialog = RangeWidget(self.main_widget.var_widget, accepted, title="Extend by zeros",
                                 label_text="Set x range values. Values outside the range of spectrum "
                                            "will be set up to zero. Spacing is calculated as average "
                                            "of the spectrum spacing (spacing will be the same for "
                                            "evenly spaced spectrum:",
                                 parent=self)

    def baseline_correct(self):

        if len(self.selectedIndexes()) == 0:
            return

        def accepted():
            # Logger.status_message("Baseline correcting...")
            x0, x1 = rng_dialog.returned_range
            try:
                items = []
                for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                                skip_childs_in_selected_groups=False):
                    item.baseline_correct_no_update(x0, x1)
                    items.append(item)

                self.data_modified(items)
                self.state_changed.emit()
            except ValueError as ex:
                Logger.message(ex.__str__())
                QMessageBox.warning(self, 'Warning', ex.__str__(), QMessageBox.Ok)
                return

            Logger.status_message("Done")

        rng_dialog = RangeWidget(self.main_widget.var_widget, accepted, title="Baseline correction",
                                 label_text="Set x range values. Y values of this range will "
                                            "be averaged and subtracted from all points:",
                                 parent=self)

    def interpolate(self):

        if len(self.selectedIndexes()) == 0:
            return

        if InterpolateDialog.is_opened:
            InterpolateDialog.get_instance().activateWindow()
            InterpolateDialog.get_instance().setFocus()
            return

        interp_dialog = InterpolateDialog(self)

        if not interp_dialog.accepted:
            return

        spacing, kind = interp_dialog.spacing, interp_dialog.selected_kind

        try:
            items = []
            for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                            skip_childs_in_selected_groups=False):
                item.interpolate_no_update(spacing, kind)
                items.append(item)

            self.data_modified(items)
            self.info_modified(items)
            self.state_changed.emit()
        except ValueError as ex:
            Logger.message(ex.__str__())
            QMessageBox.warning(self, 'Warning', ex.__str__(), QMessageBox.Ok)
            return

        Logger.status_message("Done")

    def select_every_nth_item(self):

        items_count = len(self.selectedIndexes()) / 2

        if items_count == 0:
            return

        if IntIntInputDialog.is_opened:
            IntIntInputDialog.get_instance().activateWindow()
            IntIntInputDialog.get_instance().setFocus()
            return

        n, shift = 2, 0

        intintinput_dialog = IntIntInputDialog(n, shift, n_min=1, offset_min=0,
                                               title="Select every n-th item",
                                               label="Set the n value and shift value. Group items will be skipped:")
        if not intintinput_dialog.accepted:
            return
        n, shift = intintinput_dialog.returned_range

        try:
            shift = shift % n
            i = 0

            self.selecting = True

            flags = QItemSelectionModel.Select
            selection = QItemSelection()

            for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                            skip_childs_in_selected_groups=False,
                                                            clear_selection=True):
                if i % n == shift:
                    start_index = self.myModel.createIndex(item.row(), 0, item)
                    end_index = self.myModel.createIndex(item.row(), 1, item)

                    selection.select(start_index, end_index)

                i += 1
            self.selectionModel().select(selection, flags)
            # self.selecting = False

        except Exception as ex:
            Logger.message(ex.__str__())
            QMessageBox.warning(self, 'Error', ex.__str__(), QMessageBox.Ok)
        finally:
            self.selecting = False

    def fit_curve(self):

        items_count = len(self.selectedIndexes()) / 2

        if items_count == 0:
            return

        selected_node = self.myModel.node_from_index(self.selectedIndexes()[0])
        # if isinstance(selected_node, SpectrumItemGroup):  # TODO>>>>
        #     return

        def accepted():
            self.import_spectra([fit_dialog.fits, fit_dialog.residuals])
            self.state_changed.emit()

        fit_dialog = FitWidget(self.main_widget.var_widget, accepted, selected_node, parent=self)
        Console.push_variables({'fw': fit_dialog})

    def set_style(self):

        items_count = len(self.selectedIndexes()) / 2

        if items_count == 0:
            return

        selected_node = self.myModel.node_from_index(self.selectedIndexes()[0])
        if isinstance(selected_node, SpectrumItemGroup):
            selected_node = selected_node[0]  # select spectrum

        def accepted():
            items = []
            for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                            skip_childs_in_selected_groups=False):
                color = None
                if not style_widget.cbColor.isChecked() and style_widget.color is not None:
                    color = style_widget.color.name(QColor.HexRgb)  # in rgb format

                line_alpha = int(style_widget.sbAlpha.value())

                line_width = None if style_widget.cbLineWidth.isChecked() else float(
                    style_widget.dsbLineWidth.value())
                line_type = None if style_widget.cbLineType.isChecked() else \
                    style_widget.line_types[style_widget.combLineType.currentIndex()]['index']

                symbol = style_widget.symbol_types[style_widget.combSymbol.currentIndex()]['sym']

                sym_brush_color = None
                if not style_widget.cbSymBrushDefault.isChecked() and style_widget.sym_brush_color is not None:
                    # style_widget.sym_brush_color.setAlpha(style_widget.sbSymBrushAlpha.value())
                    sym_brush_color = style_widget.sym_brush_color.name(QColor.HexRgb)  # in rgb format

                sym_fill_color = None
                if not style_widget.cbSymFillDefault.isChecked() and style_widget.sym_fill_color is not None:
                    # style_widget.sym_fill_color.setAlpha(style_widget.sbSymFillAlpha.value())
                    sym_fill_color = style_widget.sym_fill_color.name(QColor.HexRgb)  # in rgb format

                sym_brush_alpha = int(style_widget.sbSymBrushAlpha.value())
                sym_fill_alpha = int(style_widget.sbSymFillAlpha.value())
                symbol_size = float(style_widget.dsbSymSize.value())

                plot_legend = style_widget.cbPlotLegend.isChecked()

                if not hasattr(item, 'plot_legend'):
                    setattr(item, 'plot_legend', plot_legend)

                item.color = color
                item.line_width = line_width
                item.line_type = line_type
                item.plot_legend = plot_legend

                setattr(item, 'symbol', symbol)
                setattr(item, 'symbol_brush', sym_brush_color)
                setattr(item, 'symbol_fill', sym_fill_color)
                setattr(item, 'symbol_size', symbol_size)

                setattr(item, 'line_alpha', line_alpha)
                setattr(item, 'sym_brush_alpha', sym_brush_alpha)
                setattr(item, 'sym_fill_alpha', sym_fill_alpha)

                items.append(item)

            self.check_changed(items, True)
            self.state_changed.emit()

        style_widget = StyleWidget(self.main_widget.var_widget, accepted, selected_node, parent=self)

    def rename(self):

        items_count = len(self.selectedIndexes()) / 2

        if items_count == 0:
            return

        if RenameDialog.is_opened:
            RenameDialog.get_instance().activateWindow()
            RenameDialog.get_instance().setFocus()
            return

        expression, offset, c_mult_factor = Settings.last_rename_expression, 0, 1

        rename_dialog = RenameDialog(expression, offset,
                                     last_rename_take_name_from_list=Settings.last_rename_take_name_from_list)
        if not rename_dialog.accepted:
            return

        if rename_dialog.is_renaming_by_expression:
            expression, offset, c_mult_factor = rename_dialog.result
        else:
            import csv
            splitted_list = csv.reader([rename_dialog.list], doublequote=True, skipinitialspace=True,
                                       delimiter=',').__next__()

            if len(splitted_list) == 0:
                return

        try:
            i = 0
            for item in self.myModel.iterate_selected_items(skip_groups=True,
                                                            skip_childs_in_selected_groups=False):
                name = item.name
                if rename_dialog.is_renaming_by_expression:
                    name = rename(expression, name, float(offset) * float(c_mult_factor))
                    offset += 1
                else:
                    try:
                        name = splitted_list[i].strip()
                        i += 1
                    except IndexError:
                        pass
                item.name = name

            self.redraw_spectra.emit()
            self.update_view()
            self.state_changed.emit()

            Settings.last_rename_expression = expression
            Settings.last_rename_take_name_from_list = not rename_dialog.is_renaming_by_expression
            Settings.save()
        except Exception as ex:
            Logger.message(ex.__str__())
            QMessageBox.warning(self, 'Error', ex.__str__(), QMessageBox.Ok)

    def context_menu(self, pos):
        """Creates a context menu in a TreeWidget."""

        # pos is local position on QTreeWidget
        # cursor.pos() is position on screen

        item = self.myModel.node_from_index(self.indexAt(pos))

        # print(pos, item.text(0) if item is not None else "None")

        menu = QMenu()

        sel_it_menu = QMenu("With Selected Items")

        check_selected_items = QAction("Check Items (Ctrl + Q)", self)
        # check_selected_items.setShortcut(QKeySequence(Qt.Key_Control, Qt.Key_D))

        sel_it_menu.addAction(check_selected_items)
        check_selected_items.triggered.connect(self.check_selected_items)

        uncheck_selected_items = sel_it_menu.addAction("Uncheck Items (Ctrl + W)")
        uncheck_selected_items.triggered.connect(self.uncheck_selected_items)

        select_every_nth = sel_it_menu.addAction("Select Every n-th Item (Ctrl + D)")
        select_every_nth.triggered.connect(self.select_every_nth_item)

        rename = sel_it_menu.addAction("Rename Items (Ctrl + R)")
        rename.triggered.connect(self.rename)

        sel_it_menu.addSeparator()

        cut = sel_it_menu.addAction("Cut (Ctrl + T)")
        cut.triggered.connect(self.cut)

        baseline_correct = sel_it_menu.addAction("Baseline Correct (Ctrl + B)")
        baseline_correct.triggered.connect(self.baseline_correct)

        normalize = sel_it_menu.addAction("Normalize (Ctrl + N)")
        normalize.triggered.connect(self.normalize)

        extend_by_zeros = sel_it_menu.addAction("Extend By Zeros")
        extend_by_zeros.triggered.connect(self.extend_by_zeros)

        interpolate = sel_it_menu.addAction("Interpolate")
        interpolate.triggered.connect(self.interpolate)

        fit_curve = sel_it_menu.addAction("Fit Curve")
        fit_curve.triggered.connect(self.fit_curve)

        sel_it_menu.addSeparator()

        set_style = sel_it_menu.addAction("Set Style")
        set_style.triggered.connect(self.set_style)

        copy_to_clipboard = sel_it_menu.addAction("Copy Items To Clipboard (Ctrl + C)")
        copy_to_clipboard.triggered.connect(self.copy_selected_items_to_clipboard)

        export = sel_it_menu.addAction("Export Items As")
        export.triggered.connect(self.export_selected_items_as)

        # sort group menu

        sort_group_menu = QMenu("Sort Group")

        sort_group_ascending = QAction("Ascending", self)
        sort_group_descending = QAction("Descending", self)

        sort_group_menu.addAction(sort_group_ascending)
        sort_group_menu.addAction(sort_group_descending)
        sort_group_ascending.triggered.connect(lambda: self.sort_selected_group(ascending=True))
        sort_group_descending.triggered.connect(lambda: self.sort_selected_group(ascending=False))

        # sort top level items menu

        sort_top_level_items_menu = QMenu("Sort Top-level Items")

        sort_tli_ascending = QAction("Ascending", self)
        sort_tli_descending = QAction("Descending", self)

        sort_top_level_items_menu.addAction(sort_tli_ascending)
        sort_top_level_items_menu.addAction(sort_tli_descending)
        sort_tli_ascending.triggered.connect(lambda: self.sort_tree_view(ascending=True))
        sort_tli_descending.triggered.connect(lambda: self.sort_tree_view(ascending=False))

        # sort all items also withing all groups

        sort_all_menu = QMenu("Sort All Items and Groups")

        sort_all_ascending = QAction("Ascending", self)
        sort_all_descending = QAction("Descending", self)

        sort_all_menu.addAction(sort_all_ascending)
        sort_all_menu.addAction(sort_all_descending)
        sort_all_ascending.triggered.connect(lambda: self.sort_tree_view(sort_groups=True, ascending=True))
        sort_all_descending.triggered.connect(lambda: self.sort_tree_view(sort_groups=True, ascending=False))

        # clicked into blank
        if item is self.myModel.root:
            create_group = menu.addAction("Create a New Group")
            create_group.triggered.connect(self.create_group)

            uncheck_all = menu.addAction("Uncheck All Items")
            uncheck_all.triggered.connect(self.uncheck_all)

            menu.addSeparator()

            menu.addMenu(sort_top_level_items_menu)
            menu.addMenu(sort_all_menu)
        else:
            # Clicked on item
            if isinstance(item, SpectrumItem):
                add_to_group = menu.addAction("Move Items to New Group")
                add_to_group.triggered.connect(lambda: self.add_selected_items_to_group(copy=False))

                copy_to_group = menu.addAction("Copy Items to New Group")
                copy_to_group.triggered.connect(lambda: self.add_selected_items_to_group(copy=True))

                uncheck_all = menu.addAction("Uncheck All Items")
                uncheck_all.triggered.connect(self.uncheck_all)

                menu.addMenu(sel_it_menu)

                menu.addSeparator()

            # Clicked on group
            else:
                ungroup = menu.addAction("Ungroup")
                ungroup.triggered.connect(self.ungroup_selected_group)

                uncheck_all = menu.addAction("Uncheck All Items")
                uncheck_all.triggered.connect(self.uncheck_all)

                menu.addMenu(sel_it_menu)

                menu.addSeparator()

                menu.addMenu(sort_group_menu)

            menu.addMenu(sort_top_level_items_menu)
            menu.addMenu(sort_all_menu)

            menu.addSeparator()

            delete_items = menu.addAction("Remove")
            delete_items.triggered.connect(self.delete_selected_items)

        menu.addSeparator()

        paste_from_clipboard = menu.addAction("Paste Data From Clipboard (Ctrl + V)")
        paste_from_clipboard.triggered.connect(self.paste_from_clipboard)

        cursor = QCursor()

        menu.exec_(cursor.pos())

    def keyPressEvent(self, e):
        # e is QKeyEvent object
        #
        # print(e.key())
        # return

        if e.key() == 16777223:  # pressed delete
            self.delete_selected_items()

        if e.key() == 67 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + C
            self.copy_selected_items_to_clipboard()

        if e.key() == 86 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + V
            self.paste_from_clipboard()

        if e.key() == 65 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + A
            self.selectAll()

        if e.key() == 66 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + B
            self.baseline_correct()

        if e.key() == 82 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + R
            self.rename()

        if e.key() == 78 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + N
            self.normalize()

        if e.key() == 81 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + Q
            self.check_selected_items()

        if e.key() == 87 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + W
            self.uncheck_selected_items()

        if e.key() == 84 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + T
            self.cut()

        if e.key() == 68 and e.modifiers() == Qt.ControlModifier:
            # Ctrl + D
            self.select_every_nth_item()

    def import_project(self, generic_item):
        if generic_item is None:
            return

        if not hasattr(generic_item, 'children'):
            raise ValueError("Argument 'generic_item' must have an attribute 'children'.")

        num_items = len(generic_item.children)

        if num_items == 0:
            return

        self.myModel.root.children += generic_item.children

        for child in self.myModel.root.children:  # set the parent to be the current root
            child.parent = self.myModel.root

        del generic_item  # destroy the original object

        if num_items > 0:
            # update names and view
            self.myModel.insertRows(self.myModel.root.__len__(), num_items, QModelIndex())
            self.setup_info()
            self.save_state()

        self.redraw_spectra.emit()

    def add_to_list(self, items):
        """
        Copies all spectra and import them to the treewidget
        :param spectra: input parameter can be single spectrum or
        spectrumlist object or simple list of spectrast of spectra.
        """

        if not isinstance(items, list):
            items = [items.__copy__()]

        self.import_spectra(items)

    def import_spectra(self, spectra):
        if spectra is None:
            return

        if not isinstance(spectra, list):
            raise ValueError("Argument spectra must be type of list.")

        if len(spectra) == 0:
            return

        add_rows_count = 0

        for node in spectra:
            if isinstance(node, SpectrumItem):
                # child = SpectrumItem(node, node.name, '', parent=self.myModel.root)
                node.setParent(self.myModel.root)
                add_rows_count += 1
            elif isinstance(node, Spectrum):
                SpectrumItem.from_spectrum(node, parent=self.myModel.root)
                add_rows_count += 1
            if isinstance(node, list):  # for backward compatibility
                group_item = SpectrumItemGroup(name=node[0].group_name, info='', parent=self.myModel.root)
            elif isinstance(node, SpectrumList):  # list of spectra
                group_item = SpectrumItemGroup(name=node.name, info='', parent=self.myModel.root)
            else:
                continue

            add_rows_count += 1

            for sp in node:
                if isinstance(sp, SpectrumItem):
                    sp.setParent(group_item)
                else:
                    SpectrumItem.from_spectrum(sp, parent=group_item)

        if add_rows_count > 0:
            # update view and names
            self.myModel.insertRows(self.myModel.root.__len__(), add_rows_count, QModelIndex())
            self.setup_info()
            self.save_state()

    def parse_XML_Spreadsheet(self, byte_data):

        Logger.status_message("Parsing XML Spreadsheet from clipboard Mime data...")
        try:
            # get bytes from QMimeData, remove the last character and convert to string
            xml_data = byte_data[:-1].decode("utf-8")

            spectra = parse_XML_Spreadsheet(xml_data)
            self.import_spectra(spectra)
        except Exception as ex:
            Logger.message("Error reading XML Spreadsheet from clipboard Mime data.\n{}".format(ex.__str__()))
            return

        Logger.status_message("Done")

    def import_files(self, filepaths):

        # if the dropped file is a project file, load it
        if filepaths[0].lower().endswith(Settings.PROJECT_EXTENSION):
            self.main_widget.open_project(filepaths[0], open_dialog=False)
            return

        spectra, _ = parse_files(filepaths)
        self.import_spectra(spectra)

    def load_kinetics(self, dir_name, spectra_dir_name='spectra', times_fname='times.txt', blank_spectrum='blank.dx', dt=None,
                      b_corr=None, cut=None, corr_to_zero_time=True):
        """Given a directory name that contains folders of individual experiments, it loads all kinetics.
           each experiment folder must contain folder spectra (or defined in spectra_dir_name arg.)
            if blank is given, it will be subtracted from all spectra, times.txt will contain
            times for all spectra, optional baseline correction and cut can be done.

        Folder structure:
            [dir_name]
                [exp1_dir]
                    [spectra]
                        01.dx (or .csv or .txt)
                        02.dx
                        ...
                    times.txt (optional)
                    blank.dx (optional)
                [exp2_dir]
                    ...
                ...
        """

        if not os.path.isdir(dir_name):
            raise ValueError(f'{dir_name}  does not exist!')

        for item in os.listdir(dir_name):
            path = os.path.join(dir_name, item)
            if not os.path.isdir(path):
                continue

            self.load_kinetic(path, spectra_dir_name=spectra_dir_name, times_fname=times_fname, blank_spectrum=blank_spectrum,
                         dt=dt, b_corr=b_corr, cut=cut, corr_to_zero_time=corr_to_zero_time)

    def load_kinetic(self, dir_name, spectra_dir_name='spectra', times_fname='times.txt', blank_spectrum='blank.dx', dt=None,
                     b_corr=None, cut=None, corr_to_zero_time=True):
        """Given a directory name, it loads all spectra in dir named "spectra" - func. arg.,
        if blank is given, it will be subtracted from all spectra, times.txt will contain
        times for all spectra, optional baseline correction and cut can be done.

        Folder structure:
            [dir_name]
                [spectra]
                    01.dx
                    02.dx
                    ...
                times.txt (optional)
                blank.dx (optional)
        """

        root = self.myModel.root  # item in IPython console

        if not os.path.isdir(dir_name):
            raise ValueError(f'{dir_name}  does not exist!')

        spectra_path = os.path.join(dir_name, spectra_dir_name)

        if not os.path.isdir(spectra_path):
            raise ValueError(f'{spectra_dir_name}  does not exist in {dir_name}!')

        spectras = [os.path.join(spectra_path, filename) for filename in os.listdir(spectra_path)]

        n_items_before = root.__len__()
        self.import_files(spectras)
        n_spectra = root.__len__() - n_items_before

        self.add_items_to_group(root[n_items_before:], edit=False)  # add loaded spectra to group
        root[n_items_before].name = f'raw [{os.path.split(dir_name)[1]}]'  # set name of a group

        times = np.asarray([dt * i for i in range(n_spectra)]) if dt is not None else None
        # idx_add = 0
        group_idx = n_items_before
        blank_used = False

        # load explicit times
        times_fpath = os.path.join(dir_name, times_fname)
        if times is None and os.path.isfile(times_fpath):
            self.import_files(times_fpath)
            # idx_add += 1
            times = root[-1].data[:, 0].copy()
            if corr_to_zero_time:
                times -= times[0]

        if times is not None:
            root[group_idx].set_names(times)

        # load blank spectrum if available
        blank_fpath = os.path.join(dir_name, blank_spectrum)
        if os.path.isfile(blank_fpath):
            last_idx = root.__len__() - 1
            self.import_files(blank_fpath)
            self.add_to_list(root[group_idx] - root[last_idx + 1])
            if times is not None:
                root[-1].set_names(times)
            blank_used = True

        corr_idx = -1 if blank_used else group_idx

        if b_corr is not None:
            root[corr_idx].baseline_correct(*b_corr)
            root[corr_idx].name += 'bcorr'
        if cut is not None:
            root[corr_idx].cut(*cut)
            root[corr_idx].name += 'cut'







import os
import io
import numpy as np
import csv
from spectramanipulator.spectrum import Spectrum


def list_to_files(list_of_spectra, dir_path, extension, include_group_name=True, include_header=True,
                  delimiter='\t', decimal_sep='.', x_data_name='Wavelength / nm'):

    _list_to_stream(list_of_spectra, include_group_name, include_header, delimiter,
                             decimal_sep, True, dir_path, extension, x_data_name)


def list_to_string(list_of_spectra, include_group_name=True, include_header=True, delimiter='\t',
                   decimal_sep='.', x_data_name='Wavelength / nm'):

    return _list_to_stream(list_of_spectra, include_group_name, include_header, delimiter,
                                    decimal_sep, False, x_data_name=x_data_name)


def _list_to_stream(list_of_spectra, include_group_name=True, include_header=True, delimiter='\t',
                    decimal_sep='.', save_to_file=False, dir_path=None, extension=None,
                    x_data_name='Wavelength / nm'):

    # this generator iterates the ndarray data and yield returns a list of formatted row for csv writer
    def iterate_data(iterable):
        for row in iterable:
            yield [str(num).replace('.', decimal_sep) for num in row]

    if not isinstance(list_of_spectra, list):
        raise ValueError("Argument 'list_of_spectra' must be type of list.")

    buffer = ""

    dialect = csv.excel
    dialect.delimiter = delimiter
    dialect.lineterminator = '\n'
    dialect.quoting = csv.QUOTE_MINIMAL

    for i, node in enumerate(list_of_spectra):

        if save_to_file:
            name = node.name if isinstance(node, Spectrum) else node[0].group_name
            filepath = os.path.join(dir_path,
                                    (name if name != '' else 'Untitled{}'.format(i)) + extension)
            w = csv.writer(open(filepath, 'w', encoding='utf-8'), dialect=dialect)
        else:
            stream = io.StringIO('')
            w = csv.writer(stream, dialect=dialect)

        # export as group
        if isinstance(node, list):
            if len(node) == 0:
                continue

            if include_group_name:
                w.writerow([node[0].group_name])

            # add header if it is user defined
            if include_header:
                w.writerow([x_data_name] + [sp.name for sp in node])

            # add row of wavelength, then, we will transpose the matrix, otherwise, we would have to reshape
            matrix = node[0].data[:, 0]

            # add absorbance data to matrix from all exported spectra
            for sp in node:
                if sp.length() != node[0].length():
                    raise ValueError(
                        f"Spectra '{node[0].name}' and '{sp.name}' in group '{node[0].group_name}'"
                        f" have not the same length (dimension). Unable to export.")
                if sp.data[0, 0] != node[0].data[0, 0] or sp.data[-1, 0] != node[0].data[-1, 0]:
                    raise ValueError(
                        f"Spectra '{node[0].name}' and '{sp.name}' in group '{node[0].group_name}'"
                        f" have not the same length (dimension). Unable to export.")
                matrix = np.vstack((matrix, sp.data[:, 1]))

            matrix = matrix.T  # transpose

            # write matrix
            w.writerows(iterate_data(matrix))

        # export as single spectrum file
        if isinstance(node, Spectrum):
            if include_header:
                w.writerow([x_data_name, node.name])

            w.writerows(iterate_data(node.data))

        if not save_to_file:
            buffer += stream.getvalue() + '\n'
            stream.close()

    if buffer != "":
        # return ret_buffer and remove the 2 new line characters that are at the end
        return buffer[:-2]




from PyQt5.QtWidgets import QComboBox, QApplication
from PyQt5.QtGui import QStandardItem, QStandardItemModel
from PyQt5.QtCore import Qt, pyqtSignal

# from ..logger import Logger


class ComboBoxCB(QComboBox):

    check_changed = pyqtSignal(list)  # list of abbreviations of checked items

    def __init__(self, parent=None, items: dict = None):
        """items is dictionary that uses short names as keys and longer description as values"""
        super(ComboBoxCB, self).__init__(parent)

        self.setEditable(True)

        self.cb_items = []  # list of checkboxes
        self.items = {} if items is None else items
        self.model = QStandardItemModel(1, 1)
        self.setModel(self.model)
        self.model.dataChanged.connect(self.data_changed)

        self.setting_check_state = False

    def set_check_state(self, i, checked=True):
        if i >= len(self.cb_items):
            return

        self.setting_check_state = True

        self.cb_items[i].setData(Qt.Checked if checked else Qt.Unchecked, Qt.CheckStateRole)

        self.setting_check_state = False

    def set_data(self, items: dict):
        self.cb_items = []
        self.items = items
        self.model.setRowCount(len(self.items))

        for i, (key, value) in enumerate(self.items.items()):
            item = QStandardItem(f'{key}: {value}')

            item.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
            item.setData(Qt.Unchecked, Qt.CheckStateRole)

            self.cb_items.append(item)

            self.model.setItem(i, 0, item)

        self.set_text("")

    def update_text(self):
        self.data_changed(None, None, None, None)

    def data_changed(self, index1, index2, roles, p_int=None):
        if self.setting_check_state:
            return

        text = ''
        if len(self.cb_items) == len(self.items):
            key_list = [key for i, (key, value) in enumerate(self.items.items()) if
                        self.cb_items[i].checkState() == Qt.Checked]
            text = ', '.join(key_list)
            self.check_changed.emit(key_list)

        self.set_text(text)
        # Logger.debug('data_changed called')

    def set_text(self, text):
        l_edit = self.lineEdit()
        l_edit.setText(text)
        l_edit.setReadOnly(True)


if __name__ == '__main__':
    import sys

    app = QApplication(sys.argv)
    cb = ComboBoxCB(None, ['a', 'b', 'c'])
    cb.show()
    sys.exit(app.exec_())



from PyQt5 import QtWidgets

from PyQt5.QtCore import Qt, QEvent
from .gui_fit_widget import Ui_Form

from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import QMessageBox, QLineEdit, QCheckBox, QFileDialog
import numpy as np

from .checkbox_rc import CheckBoxRC

from spectramanipulator.settings import Settings
from spectramanipulator.logger import Logger
from spectramanipulator.treeview.item import SpectrumItem, SpectrumItemGroup
from spectramanipulator.console import Console
from spectramanipulator.spectrum import fi
from spectramanipulator.dialogs.trust_reg_refl_option_dialog import TrustRegionReflOptionDialog
import spectramanipulator
from .param_settings_dialog import ParamSettingsDialog
from .dial_lineedit import DialLineEdit
from .mylineedit import MyLineEdit

import pyqtgraph as pg

# from user_namespace import UserNamespace

from lmfit import Minimizer, Parameters
from scipy.integrate import odeint

import spectramanipulator.fitting.fitmodels as fitmodels

import inspect
from ..general_model import GeneralModel

from ..plotwidget import PlotWidget
from .fitresult import FitResult
from ..fitting.task_fit import TaskFit

from ..utils.syntax_highlighter import KineticModelHighlighter

import glob
import os
import time

import sys

from ..misc import int_default_color_scheme
from .combobox_cb import ComboBoxCB
from copy import deepcopy

import math

def is_nan_or_inf(value):
    return math.isnan(value) or math.isinf(value)


class FitWidget(QtWidgets.QWidget, Ui_Form):
    # static variables
    is_opened = False
    _instance = None

    # maximum number of parameters
    max_param_count = 25

    def __init__(self, dock_widget, accepted_func, node: [SpectrumItem, SpectrumItemGroup] = None, parent=None):
        super(FitWidget, self).__init__(parent)

        # if FitWidget._instance is not None:
        #     PlotWidget._instance.removeItem(FitWidget._instance.lr)

        self.setupUi(self)

        self.dock_widget = dock_widget
        self.accepted_func = accepted_func

        self.fitted_params = None
        self.covariance_matrix = None
        self.errors = None

        # if spectrum is None - we have just open the widget for function plotting
        self.node = None
        if node is not None:
            # could be spectrum or a SpectrumItemGroup for multi and batch fitting
            self.node = [node] if len(node.children) == 0 else node.children

        self.lr = None  # linear region item

        # self.current_model = None  # actual model, it could be general model or predefined model
        self._general_model = None
        self.saving_general_model = False

        # setup general model option fields

        self.cbVarProAmps.name = "varpro"
        self.cbVarProAmps.toggled.connect(self.model_option_changed)
        self.cbVarProIntercept.name = "fit_intercept_varpro"
        self.cbVarProIntercept.toggled.connect(self.model_option_changed)

        self._predefined_model = None
        self.general_model_params = None
        self.fit_result = None
        self.t_fit = None

        self.fits = SpectrumItemGroup(name='Fits')
        self.residuals = SpectrumItemGroup(name='Residuals')

        self.leX0.lb = self.leX0
        self.leX0.ub = self.leX1
        self.leX1.lb = self.leX0
        self.leX1.ub = self.leX1

        self.leX0.srb = self.btnShowRegion
        self.leX1.srb = self.btnShowRegion
        self.btnShowRegion.lb = self.leX0
        self.btnShowRegion.ub = self.leX1

        self.btnShowRegion.toggled.connect(self.show_region_checked_changed)
        self.btnShowRegion.setChecked(True)

        # update region when the focus is lost and when the user presses enter
        self.leX0.returnPressed.connect(self.update_region)
        self.leX1.returnPressed.connect(self.update_region)
        self.leX0.focus_lost.connect(self.update_region)
        self.leX1.focus_lost.connect(self.update_region)

        self.update_region_text_values(self.leX0, self.leX1)

        self.btnFit.clicked.connect(self.fit)
        self.btnSimulateModel.clicked.connect(self.simulate_model)
        self.btnPrintReport.clicked.connect(self.print_report)
        self.btnClearPlot.clicked.connect(self.clear_plot)
        self.btnOK.clicked.connect(self.accept)
        self.btnCancel.clicked.connect(self.reject)
        self.cbGenModels.currentIndexChanged.connect(self.cbGenModel_changed)
        self.btnBuildModel.clicked.connect(self.build_gen_model)
        self.cbShowBackwardRates.stateChanged.connect(self.cbShowBackwardRates_checked_changed)
        self.btnSaveCustomModel.clicked.connect(self.save_general_model_as)
        self.sbSpeciesCount.valueChanged.connect(lambda: self.n_spec_changed())
        self.cbShowResiduals.toggled.connect(self._replot)
        self.cbFitBlackColor.toggled.connect(self._replot)
        self.tbAlgorithmSettings.clicked.connect(self.tbAlgorithmSettings_clicked)

        # predefined kinetic model setup

        self.ppteScheme_highlighter = KineticModelHighlighter(self.pteScheme.document())
        self.scrollAreaPredefinedModel.viewport().installEventFilter(self)
        self.scrollAreaCustomModel.viewport().installEventFilter(self)

        self.model_options_grid_layout = QtWidgets.QGridLayout(self)
        self.pred_model_VLayout = QtWidgets.QVBoxLayout(self)
        self.general_model_VLayout = QtWidgets.QVBoxLayout(self)
        self.scrollAreaPredefinedModelContent.setLayout(self.pred_model_VLayout)
        self.scrollAreaCustomModelContent.setLayout(self.general_model_VLayout)

        self.pred_model_VLayout.addLayout(self.model_options_grid_layout)
        self.pred_model_indep_params_layout = self.create_params_layout()
        self.general_model_indep_params_layout = self.create_params_layout()
        self.pred_model_VLayout.addLayout(self.pred_model_indep_params_layout)
        self.general_model_VLayout.addLayout(self.general_model_indep_params_layout)

        self.cbPredefModelDepParams = ComboBoxCB(self)
        self.cbPredefModelDepParams_changing = False
        self.cbPredefModelDepParams.check_changed.connect(self.cbPredefModelDepParams_check_changed)

        self.cbGeneralModelDepParams = ComboBoxCB(self)
        self.cbGeneralModelDepParams_changing = False
        self.cbGeneralModelDepParams.check_changed.connect(self.cbGeneralModelDepParams_check_changed)

        hbox = QtWidgets.QHBoxLayout(self)
        hbox2 = QtWidgets.QHBoxLayout(self)
        hbox.addWidget(QtWidgets.QLabel('Selected experiment-dependent parameters:'))
        hbox2.addWidget(QtWidgets.QLabel('Selected experiment-dependent parameters:'))
        hbox.addWidget(self.cbPredefModelDepParams)
        hbox2.addWidget(self.cbGeneralModelDepParams)
        btnExperimentParamsSettings_predefined = QtWidgets.QToolButton()
        btnExperimentParamsSettings_genearal = QtWidgets.QToolButton()
        btnExperimentParamsSettings_predefined.setText('...')
        btnExperimentParamsSettings_genearal.setText('...')
        btnExperimentParamsSettings_predefined.clicked.connect(self.exp_dep_param_sett_clicked)
        btnExperimentParamsSettings_genearal.clicked.connect(self.exp_dep_param_sett_clicked)
        hbox.addWidget(btnExperimentParamsSettings_predefined)
        hbox2.addWidget(btnExperimentParamsSettings_genearal)
        self.pred_model_VLayout.addLayout(hbox)
        self.general_model_VLayout.addLayout(hbox2)

        self.tab_widget_pred_model = QtWidgets.QTabWidget(self)
        self.pred_model_VLayout.addWidget(self.tab_widget_pred_model)

        self.tab_widget_general_model = QtWidgets.QTabWidget(self)
        self.general_model_VLayout.addWidget(self.tab_widget_general_model)

        self.pred_species_hlayouts = []
        self.pred_model_dep_param_layouts = []
        self.pred_model_bounds = {'lb': [], 'ub': []}

        self.general_species_hlayouts = []
        self.general_model_dep_param_layouts = []
        self.general_model_bounds = {'lb': [], 'ub': []}

        if self.node is not None:
            for spectrum in self.node:
                widget, species_hlayout, params_layout, lower_bound, upper_bound, show_region_btn = self.create_tab_widget(self.tab_widget_pred_model)
                self.pred_species_hlayouts.append(species_hlayout)
                self.pred_model_dep_param_layouts.append(params_layout)
                self.pred_model_bounds['lb'].append(lower_bound)
                self.pred_model_bounds['ub'].append(upper_bound)

                lower_bound.lb = lower_bound
                lower_bound.ub = upper_bound
                upper_bound.lb = lower_bound
                upper_bound.ub = upper_bound

                lower_bound.srb = show_region_btn
                upper_bound.srb = show_region_btn

                lower_bound.returnPressed.connect(self.update_region)
                upper_bound.returnPressed.connect(self.update_region)
                lower_bound.focus_lost.connect(self.update_region)
                upper_bound.focus_lost.connect(self.update_region)
                show_region_btn.toggled.connect(self.show_region_checked_changed)
                show_region_btn.lb = lower_bound
                show_region_btn.ub = upper_bound

                widget_2, species_hlayout_2, params_layout_2, lower_bound_2, upper_bound_2, show_region_btn_2 = self.create_tab_widget(self.tab_widget_general_model)
                self.general_species_hlayouts.append(species_hlayout_2)
                self.general_model_dep_param_layouts.append(params_layout_2)
                self.general_model_bounds['lb'].append(lower_bound_2)
                self.general_model_bounds['ub'].append(upper_bound_2)

                lower_bound_2.lb = lower_bound_2
                lower_bound_2.ub = upper_bound_2
                upper_bound_2.lb = lower_bound_2
                upper_bound_2.ub = upper_bound_2

                lower_bound_2.srb = show_region_btn_2
                upper_bound_2.srb = show_region_btn_2

                lower_bound_2.returnPressed.connect(self.update_region)
                upper_bound_2.returnPressed.connect(self.update_region)
                lower_bound_2.focus_lost.connect(self.update_region)
                upper_bound_2.focus_lost.connect(self.update_region)
                show_region_btn_2.toggled.connect(self.show_region_checked_changed)
                show_region_btn_2.lb = lower_bound_2
                show_region_btn_2.ub = upper_bound_2

                self.tab_widget_pred_model.addTab(widget, spectrum.name)
                self.tab_widget_general_model.addTab(widget_2, spectrum.name)

        vals_temp = dict(params=[], lower_bounds=[], values=[], upper_bounds=[], fixed=[], errors=[])
        self.pred_param_fields = dict(exp_independent=deepcopy(vals_temp),
                                      exp_dependent=[deepcopy(vals_temp) for _ in range(len(self.node))])

        self.pred_spec_visible_fields = [[] for _ in range(len(self.node))]
        self.general_param_fields = deepcopy(self.pred_param_fields)
        self.general_spec_visible_fields = deepcopy(self.pred_spec_visible_fields)

        def add_widgets(dict_fields: dict, param_layout: QtWidgets.QGridLayout):
            for i in range(self.max_param_count):
                lb = QLineEdit()  # lower bound text field
                par = QLineEdit()  # value text field
                ub = QLineEdit()  # upper bound text field
                fix = CheckBoxRC()  # fixed checkbox
                err = QLineEdit()  # std err text field
                val = DialLineEdit(lb_val=lb.text, ub_val=ub.text)  # parameter text field

                par.setReadOnly(True)
                err.setReadOnly(True)

                # set par as an attribute to all other field to be able to access the param field from them
                lb.par = par
                val.le.par = par
                ub.par = par
                fix.par = par

                # set lower and upper bound fields as tag for fix field to be able to handle fix field changes
                fix.lb = lb
                fix.ub = ub

                par.setVisible(False)
                lb.setVisible(False)
                val.setVisible(False)
                ub.setVisible(False)
                fix.setVisible(False)
                err.setVisible(False)

                # update model param upon field change
                lb.textChanged.connect(lambda value: self.transfer_param_to_model('min', value))
                val.textChanged.connect(lambda value: self.transfer_param_to_model('value', value))
                ub.textChanged.connect(lambda value: self.transfer_param_to_model('max', value))
                fix.toggled.connect(lambda value: self.transfer_param_to_model('vary', not value))
                val.simulation_requested.connect(self._simulate)  # simulate

                # put references to dictionary
                dict_fields['params'].append(par)
                dict_fields['lower_bounds'].append(lb)
                dict_fields['values'].append(val)
                dict_fields['upper_bounds'].append(ub)
                dict_fields['fixed'].append(fix)
                dict_fields['errors'].append(err)

                # add widgets to layout
                param_layout.addWidget(par, i + 1, 0, 1, 1)
                param_layout.addWidget(lb, i + 1, 1, 1, 1)
                param_layout.addLayout(val, i + 1, 2, 1, 1)
                param_layout.addWidget(ub, i + 1, 3, 1, 1)
                param_layout.addWidget(fix, i + 1, 4, 1, 1)
                param_layout.addWidget(err, i + 1, 5, 1, 1)

        def set_tab_order(dict_fields: dict):
            for _ in range(self.max_param_count):
                for key in dict_fields.keys():
                    for first, second in zip(dict_fields[key][:-1], dict_fields[key][1:]):
                        if isinstance(first, DialLineEdit):
                            first = first.le

                        if isinstance(second, DialLineEdit):
                            second = second.le

                        self.setTabOrder(first, second)

        add_widgets(self.pred_param_fields['exp_independent'], self.pred_model_indep_params_layout)
        set_tab_order(self.pred_param_fields['exp_independent'])

        add_widgets(self.general_param_fields['exp_independent'], self.general_model_indep_params_layout)
        set_tab_order(self.general_param_fields['exp_independent'])

        for i in range(len(self.node)):
            add_widgets(self.pred_param_fields['exp_dependent'][i], self.pred_model_dep_param_layouts[i])
            set_tab_order(self.pred_param_fields['exp_dependent'][i])

            add_widgets(self.general_param_fields['exp_dependent'][i], self.general_model_dep_param_layouts[i])
            set_tab_order(self.general_param_fields['exp_dependent'][i])

            # setup species visible checkboxes
            for _ in range(self.max_param_count):
                cb_predefined_model = CheckBoxRC('')
                cb_predefined_model.exp_index = i
                cb_predefined_model.setVisible(False)
                cb_general_model = CheckBoxRC('')
                cb_general_model.exp_index = i
                cb_general_model.setVisible(False)

                cb_predefined_model.toggled.connect(self.species_visible_checkbox_toggled)
                cb_general_model.toggled.connect(self.species_visible_checkbox_toggled)

                self.pred_spec_visible_fields[i].append(cb_predefined_model)
                self.pred_species_hlayouts[i].addWidget(cb_predefined_model)

                self.general_spec_visible_fields[i].append(cb_general_model)
                self.general_species_hlayouts[i].addWidget(cb_general_model)
            self.pred_species_hlayouts[i].addSpacerItem(QtWidgets.QSpacerItem(1, 1,
                                                        QtWidgets.QSizePolicy.Expanding,
                                                        QtWidgets.QSizePolicy.Fixed))
            self.general_species_hlayouts[i].addSpacerItem(QtWidgets.QSpacerItem(1, 1,
                                                           QtWidgets.QSizePolicy.Expanding,
                                                           QtWidgets.QSizePolicy.Fixed))

        self.model_option_widgets = []  # only for predefined model

        self.setting_params = False

        # get all models from fitmodels, get classes that inherits from Model base class and sort them by name
        # and number of species
        classes = inspect.getmembers(sys.modules[fitmodels.__name__], inspect.isclass)
        tuples = filter(lambda tup: issubclass(tup[1], fitmodels.Model) and
                        (tup[1] is not fitmodels.Model
                         and tup[1] is not fitmodels.GeneralFitModel
                         and tup[1] is not fitmodels._InterceptVarProModel
                         and tup[1] is not fitmodels._FixedParametersModel), classes)
        # load models
        self.models = sorted(list(map(lambda tup: tup[1], tuples)), key=lambda cls: cls.name)
        # fill the available models combo box with model names
        self.cbModel.addItems(map(lambda m: m.name, self.models))

        self.cbModel.currentIndexChanged.connect(self.predefined_model_changed)

        self.gen_models_paths = []
        self.update_general_models()

        self.res_weights = [
            {'description': 'No weighting', 'func': lambda res, y: res},
            {'description': 'Absorbance weighting: noise \u221d 10 ^ y', 'func': lambda res, y: res * np.exp(-y * np.log(10))},
            {'description': 'Proportional weighting: noise \u221d y', 'func': lambda res, y: res / y},
        ]

        self.cbResWeighting.addItems(map(lambda m: m['description'], self.res_weights))

        # abbr is name that is needed for lmfit.fitting method
        self.methods = [
            {'name': 'Trust Region Reflective method', 'abbr': 'least_squares', 'option_dialog': TrustRegionReflOptionDialog},
            {'name': 'Levenberg–Marquardt', 'abbr': 'leastsq', 'option_dialog': None},
            {'name': 'Nelder-Mead, Simplex method (no error)', 'abbr': 'nelder', 'option_dialog': None},
            {'name': 'Differential evolution', 'abbr': 'differential_evolution', 'option_dialog': None},
            {'name': 'L-BFGS-B (no error)', 'abbr': 'lbfgsb', 'option_dialog': None},
            {'name': 'Powell (no error)', 'abbr': 'powell', 'option_dialog': None}
        ]

        self.cbMethod.addItems(map(lambda m: m['name'], self.methods))
        self.cbMethod.currentIndexChanged.connect(self.cbMethod_currentIndexChanged)
        self.cbMethod_currentIndexChanged()
        self.model_options_dict = {}

        self.accepted = False
        FitWidget.is_opened = True
        FitWidget._instance = self

        self.dock_widget.parent().resizeDocks([self.dock_widget], [250], Qt.Vertical)
        self.title_text = 'Function plotter'
        if self.node is not None:
            self.title_text = 'Fit of {}'.format(self.node[0].parent.name if len(self.node) > 0 else self.node[0].name)
        self.dock_widget.titleBarWidget().setText(self.title_text)
        self.dock_widget.setWidget(self)
        self.dock_widget.setVisible(True)

        self.setup_general_model()
        self.predefined_model_changed()
        self.cbMethod_currentIndexChanged()

        self.tab_widget_pred_model.currentChanged.connect(lambda: self.setup_fields(self._predefined_model))
        self.tab_widget_general_model.currentChanged.connect(lambda: self.setup_fields(self._general_model))

    def eventFilter(self, source, event):
        # disable scrolling from inside of QScrollAreas
        # https://stackoverflow.com/questions/46745800/disable-mouse-wheel-scroll-on-qscrollarea
        if (event.type() == QEvent.Wheel and
                (source is self.scrollAreaCustomModel.viewport() or
                 source is self.scrollAreaPredefinedModel.viewport())):
            return True
        return super(FitWidget, self).eventFilter(source, event)

    def show_region_checked_changed(self):
        sender = self.sender()

        if not sender.isChecked():
            PlotWidget.remove_linear_region()
            self.lr = None
            return

        bounds = None
        if self.node is not None:
            # find the max range for all of the data

            x0 = self.node[0].data[0, 0]
            x1 = self.node[0].data[-1, 0]

            for item in self.node[1:]:
                new_x0 = item.data[0, 0]
                new_x1 = item.data[-1, 0]
                if new_x0 < x0:
                    x0 = new_x0
                if new_x1 > x1:
                    x1 = new_x1
            x0 = -1 if is_nan_or_inf(x0) else x0
            x1 = 1 if is_nan_or_inf(x1) else x1

            bounds = (x0, x1)

        try:
            region = float(sender.lb.text()), float(sender.ub.text())
        except ValueError:
            region = None

        self.lr = PlotWidget.add_linear_region(bounds=bounds, z_value=1e8, region=region)
        self.lr.sigRegionChanged.connect(lambda: self.update_region_text_values(sender.lb, sender.ub))

    def update_region_text_values(self, lb, ub):
        x0, x1 = self.lr.getRegion()
        lb.setText("{:.4g}".format(x0))
        ub.setText("{:.4g}".format(x1))

    def update_region(self):
        if self.lr is None:
            return

        sender = self.sender()

        lb = sender.lb
        ub = sender.ub

        if not sender.srb.isChecked():
            return

        try:
            x0, x1 = float(lb.text()), float(ub.text())
            self.lr.setRegion((x0, x1))
        except ValueError:
            pass

    def create_tab_widget(self, tab_widget):
        widget = QtWidgets.QWidget(tab_widget)
        vl = QtWidgets.QVBoxLayout(widget)
        vl.addWidget(QtWidgets.QLabel('Species visible:'))
        species_hlayout = QtWidgets.QHBoxLayout(widget)  # species visible layout
        vl.addLayout(species_hlayout)
        # vl.addWidget(QtWidgets.QLabel('Experiment-dependent parameters:'))
        params_layout = self.create_params_layout(widget)
        vl.addLayout(params_layout)
        hlayout_range = QtWidgets.QHBoxLayout(widget)
        lower_bound = MyLineEdit(widget)
        upper_bound = MyLineEdit(widget)
        show_region_btn = QtWidgets.QPushButton("Show Region", widget)
        show_region_btn.setCheckable(True)
        show_region_btn.setChecked(False)
        vl.addWidget(QtWidgets.QLabel("Fit range:"))
        hlayout_range.addWidget(QtWidgets.QLabel("from"))
        hlayout_range.addWidget(lower_bound)
        hlayout_range.addWidget(QtWidgets.QLabel("to"))
        hlayout_range.addWidget(upper_bound)
        hlayout_range.addWidget(show_region_btn)
        vl.addLayout(hlayout_range)
        vl.addSpacerItem(QtWidgets.QSpacerItem(1, 1,
                         QtWidgets.QSizePolicy.Fixed,
                         QtWidgets.QSizePolicy.Expanding))

        widget.setLayout(vl)

        return widget, species_hlayout, params_layout, lower_bound, upper_bound, show_region_btn

    def create_params_layout(self, parent=None):
        params_layout_template = QtWidgets.QGridLayout(self if parent is None else parent)
        param_label = QtWidgets.QLabel('Param')
        lb_label = QtWidgets.QLabel('Lower\nBound')
        val_label = QtWidgets.QLabel('\u2264 Value \u2264')  # <= Value <=
        ub_label = QtWidgets.QLabel('Upper\nBound')
        fix_label = QtWidgets.QLabel('Fixed')
        err_label = QtWidgets.QLabel('Error')
        param_label.setAlignment(Qt.AlignCenter)
        lb_label.setAlignment(Qt.AlignCenter)
        val_label.setAlignment(Qt.AlignCenter)
        ub_label.setAlignment(Qt.AlignCenter)
        fix_label.setAlignment(Qt.AlignCenter)
        err_label.setAlignment(Qt.AlignCenter)
        params_layout_template.addWidget(param_label, 0, 0, 1, 1)
        params_layout_template.addWidget(lb_label, 0, 1, 1, 1)
        params_layout_template.addWidget(val_label, 0, 2, 1, 1)
        params_layout_template.addWidget(ub_label, 0, 3, 1, 1)
        params_layout_template.addWidget(fix_label, 0, 4, 1, 1)
        params_layout_template.addWidget(err_label, 0, 5, 1, 1)
        params_layout_template.setVerticalSpacing(0)

        return params_layout_template

    def update_general_models(self):
        # curr_idx = self.cbGenModels.currentIndex() if len(self.cbGenModels) > 0 else None
        self.cbGenModels.clear()
        self.gen_models_paths = []

        module_path = os.path.abspath(spectramanipulator.__file__)  # get the absolute path of a module
        fpath = os.path.join(os.path.dirname(module_path), Settings.general_models_dir, '*.json')

        for fpath in glob.glob(fpath, recursive=True):
            fname = os.path.splitext(os.path.split(fpath)[1])[0]
            self.gen_models_paths.append(fpath)
            self.cbGenModels.addItem(fname)

    @staticmethod
    def check_state(checked):
        return Qt.Checked if checked else 0

    def save_general_model_as(self):
        if self._general_model is None:
            return

        curr_model_path = self.gen_models_paths[self.cbGenModels.currentIndex()]
        file = "Json (*.json)"

        filepath = QFileDialog.getSaveFileName(caption="Save Custom Kinetic Model",
                                               directory=curr_model_path,
                                               filter=file, initialFilter=file)
        if filepath[0] == '':
            return

        try:
            self._general_model.scheme = self.pteScheme.toPlainText()
            j = np.asarray([self._general_model.params[p].value for p in self._general_model.param_names_dict[0]['j']])
            rates = self._general_model.get_rate_values(0)

            self._general_model.set_rates(rates)
            self._general_model.initial_conditions = dict(zip(self._general_model.get_compartments(), j))
            self._general_model.save(filepath[0])
        except Exception as e:
            Logger.message(e.__str__())
            QMessageBox.critical(self, 'Saving error', 'Error, first build the model.', QMessageBox.Ok)
            return

        self.saving_general_model = True
        self.update_general_models()
        # set saved model as current index
        for k, fpath in enumerate(self.gen_models_paths):
            fname = os.path.splitext(os.path.split(fpath)[1])[0]
            if fname == os.path.splitext(os.path.split(filepath[0])[1])[0]:
                self.cbGenModels.setCurrentIndex(k)
                break

        self.saving_general_model = False

    def cbGenModel_changed(self):
        if self._general_model is None or self.saving_general_model:
            return

        self._general_model.load_from_file(self.gen_models_paths[self.cbGenModels.currentIndex()])
        self.pteScheme.setPlainText(self._general_model.general_model.scheme)
        self.general_model_setup_dep_params()
        self.setup_fields(self._general_model)

    def cbShowBackwardRates_checked_changed(self):
        if self._general_model is None:
            return

        self._general_model.update_model_options(show_backward_rates=self.cbShowBackwardRates.isChecked())
        self.general_model_setup_dep_params()
        self.setup_fields(self._general_model)

    def build_gen_model(self):
        self._general_model.load_from_scheme(self.pteScheme.toPlainText())
        self.general_model_setup_dep_params()
        self.setup_fields(self._general_model)

    def cbGeneralModelDepParams_check_changed(self, checked_abbrs):
        if self.cbGeneralModelDepParams_changing:
            return
        Logger.debug('cbGeneralModelDepParams_check_changed')

        self._general_model.update_model_options(exp_dep_params=checked_abbrs)
        self.setup_fields(self._general_model)

    def cbPredefModelDepParams_check_changed(self, checked_abbrs):
        if self.cbPredefModelDepParams_changing:
            return
        Logger.debug('cbPredefModelDepParams_check_changed')

        self._predefined_model.update_model_options(exp_dep_params=set(checked_abbrs))
        self.setup_fields(self._predefined_model)

    def species_visible_checkbox_toggled(self, checked):
        model = self._predefined_model if self.tabWidget.currentIndex() == 0 else self._general_model

        if model is None or self.setting_params:
            return

        cb = self.sender()
        Logger.debug('species_visible_checkbox_toggled', cb.exp_index, cb.text())

        if cb.right_button_pressed:
            model.set_all_spec_visible(checked, cb.text())
        else:
            model.spec_visible[cb.exp_index][cb.text()] = checked

        model.update_model_options()
        self.setup_fields()

    def transfer_param_to_model(self, param_type: str, value):
        """also handles enabling of lower and upper text fields"""
        if self.setting_params:
            return

        model = self._predefined_model if self.tabWidget.currentIndex() == 0 else self._general_model

        par_name = self.sender().par.text()
        if par_name == '':
            return
        if param_type != 'vary':
            try:
                value = float(value)
            except ValueError:
                return
        else:  # handles fix toggle
            # set enabled for lower and upper bound fields
            self.sender().lb.setEnabled(value)
            self.sender().ub.setEnabled(value)

            # if right mouse button was pressed, fix/unfix all experiments params
            if self.sender().right_button_pressed:
                model.fix_all_exp_dep_params(value, par_name)
                self.setup_fields()
                return

        # print('par name:', par_name, 'type', param_type, 'value', value)
        try:
            model.params[par_name].__setattr__(param_type, value)
        except KeyError as e:
            print(e.__repr__())

    def _fill_fields(self, fields: dict, params: list):
        for i in range(self.max_param_count):
            visible = len(params) > i

            if visible:
                p = params[i]
                enabled = True
                if hasattr(p, 'enabled'):
                    enabled = p.enabled

                fields['params'][i].setText(p.name)
                fields['lower_bounds'][i].setText(f'{p.min:.4g}')
                fields['lower_bounds'][i].setEnabled(p.vary)
                fields['values'][i].setText(f'{p.value:.4g}')
                fields['values'][i].setEnabled(enabled)
                fields['upper_bounds'][i].setText(f'{p.max:.4g}')
                fields['upper_bounds'][i].setEnabled(p.vary)
                fields['fixed'][i].setChecked(not p.vary)
                fields['fixed'][i].setEnabled(enabled)
                stderr = f'{p.stderr:.4g}' if p.stderr is not None else '0'
                fields['errors'][i].setText(stderr)

            fields['params'][i].setVisible(visible)
            fields['lower_bounds'][i].setVisible(visible)
            fields['values'][i].setVisible(visible)
            fields['upper_bounds'][i].setVisible(visible)
            fields['fixed'][i].setVisible(visible)
            fields['errors'][i].setVisible(visible)

    def setup_fields(self, model=None):
        """Transfer parameters from model to all fields"""
        if model is None:
            model = self._predefined_model if self.tabWidget.currentIndex() == 0 else self._general_model

        Logger.debug('setup_field')
        self.setting_params = True  # prevent calling transfer_param_to_model and species_visible_checkbox_toggled

        if model == self._predefined_model:
            param_fields = self.pred_param_fields
            vis_fields = self.pred_spec_visible_fields
        else:
            param_fields = self.general_param_fields
            vis_fields = self.general_spec_visible_fields

        self._fill_fields(param_fields['exp_independent'], model.get_model_indep_params_list())

        model_dep_params = model.get_model_dep_params_list()
        spec_names = model.get_current_species_names()
        for i in range(len(self.node)):
            self._fill_fields(param_fields['exp_dependent'][i], model_dep_params[i])

            for j in range(self.max_param_count):
                visible = len(spec_names) > j

                if visible:
                    vis_fields[i][j].setText(spec_names[j])
                    vis_fields[i][j].setChecked(model.spec_visible[i][spec_names[j]])

                vis_fields[i][j].setVisible(visible)

        self.setting_params = False

    def exp_dep_param_sett_clicked(self):
        model = self._predefined_model if self.tabWidget.currentIndex() == 0 else self._general_model

        if model is None or self.setting_params:
            return

        param_fields = self.pred_param_fields if model == self._predefined_model else self.general_param_fields
        par_names = list(model.exp_dep_params)

        exp_dep_params = []

        for i in range(len(par_names)):  # number of parameters
            lst_vals = []
            d = dict(name=par_names[i], values=lst_vals)
            for j in range(len(self.node)):  # number of experiments
                lst_vals.append(param_fields['exp_dependent'][j]['values'][i].text())
            exp_dep_params.append(d)

        def set_result():
            for i in range(len(psd.exp_dep_params)):  # number of parameters
                for j in range(len(self.node)):  # number of experiments
                    par_value = psd.exp_dep_params[i]['values'][j]
                    par_name = param_fields['exp_dependent'][j]['params'][i].text()

                    try:
                        model.params[par_name].value = par_value
                    except KeyError as e:
                        print(e.__repr__())

            model.equal_pars = psd.equal_pars  # update equal parameter setup
            model.update_params()
            self.setup_fields()  # update fields

        psd = ParamSettingsDialog(exp_dep_params, model.equal_pars, set_result=set_result)
        psd.show()

    def general_model_setup_dep_params(self):
        self.cbGeneralModelDepParams_changing = True

        av_params = self._general_model.get_all_param_names()  # available parameters

        # set default parameters to checkbox of model dependent parameters
        # check the default values
        self.cbGeneralModelDepParams.set_data(av_params)
        abbrs = list(self.cbGeneralModelDepParams.items.keys())
        for i in range(self.cbGeneralModelDepParams.items.__len__()):
            self.cbGeneralModelDepParams.set_check_state(i, abbrs[i] in self._general_model.exp_dep_params)

        self.cbGeneralModelDepParams_changing = False
        # self.cbGeneralModelDepParams_check_changed()
        self.cbGeneralModelDepParams.update_text()

    def n_spec_changed(self):
        Logger.debug('n_spec_changed')

        self.cbPredefModelDepParams_changing = True

        self._predefined_model.update_model_options(n_spec=int(self.sbSpeciesCount.value()))

        av_params = self._predefined_model.get_all_param_names()  # available parameters
        # default_params = self._predefined_model.exp_dep_params

        # set default parameters to checkbox of model dependent parameters
        # check the default values
        self.cbPredefModelDepParams.set_data(av_params)
        # default_dep_params = self._predefined_model.default_exp_dep_params()  # default values
        abbrs = list(self.cbPredefModelDepParams.items.keys())
        for i in range(self.cbPredefModelDepParams.items.__len__()):
            self.cbPredefModelDepParams.set_check_state(i, abbrs[i] in self._predefined_model.exp_dep_params)

        self.cbPredefModelDepParams_changing = False
        self.cbPredefModelDepParams.update_text()

        self.setup_fields(self._predefined_model)

    def model_option_changed(self, value):
        """also handles enabling of lower and upper text fields"""
        if self.setting_params:
            return

        model = self._predefined_model if self.tabWidget.currentIndex() == 0 else self._general_model

        opt_name = self.sender().name

        opts = {opt_name: value}

        model.update_model_options(**opts)
        self.setup_fields()

    def setup_general_model(self):
        data = [sp.data for sp in self.node] if self.node is not None else None
        self._general_model = fitmodels.GeneralFitModel(data, varpro=self.cbVarProAmps.isChecked(),
                                                        fit_intercept_varpro=self.cbVarProIntercept.isChecked())
        self.cbGenModel_changed()

    def predefined_model_changed(self):
        # initialize new model
        data = [sp.data for sp in self.node] if self.node is not None else None
        self._predefined_model = self.models[self.cbModel.currentIndex()](data, n_spec=int(self.sbSpeciesCount.value()))

        self.setup_pred_model_options()

        self.n_spec_changed()

    def setup_pred_model_options(self):
        """Add option fields that are associated with the selected model"""

        opts = self._predefined_model.model_options()

        # delete all widgets in grid_layout, use walrus operator here
        for w in self.model_option_widgets:
            self.model_options_grid_layout.removeWidget(w)  # remove widget from layout
            w.deleteLater()  # delete the widget https://stackoverflow.com/questions/10716300/removing-qwidgets-from-a-qgridlayout

        self.model_option_widgets.clear()

        for i, op in enumerate(opts):

            if op['type'] is bool:
                widget = QCheckBox(op['description'])
                widget.setChecked(op['value'])
                widget.toggled.connect(self.model_option_changed)
            else:
                widget = QLineEdit()
                widget.setText(op['value'])
                widget.textChanged.connect(self.model_option_changed)

            widget.name = op['name']  # add param name as attribute to widget

            self.model_option_widgets.append(widget)

            if op['type'] is bool:
                self.model_options_grid_layout.addWidget(widget, i, 0, 1, 2)
            else:
                qlabel = QtWidgets.QLabel(op['description'])
                self.model_option_widgets.append(qlabel)
                self.model_options_grid_layout.addWidget(qlabel, i, 0, 1, 1)
                self.model_options_grid_layout.addWidget(widget, i, 1, 1, 1)

    def clear_plot(self):
        self.fits.children.clear()
        self.residuals.children.clear()
        PlotWidget.remove_all_fits()

    def cbMethod_currentIndexChanged(self):
        dialog_cls = self.methods[self.cbMethod.currentIndex()]['option_dialog']
        if dialog_cls is None:
            self.model_options_dict = {}
            return

        self.model_options_dict = dialog_cls.default_opts()

    def tbAlgorithmSettings_clicked(self):
        def set_result():
            self.model_options_dict = dialog.options

        dialog_cls = self.methods[self.cbMethod.currentIndex()]['option_dialog']

        if dialog_cls is None:
            return

        dialog = dialog_cls(set_result=set_result, **self.model_options_dict)
        dialog.show()

    def simulate_model(self):

        # self._plot_function()
        self._simulate()
        # try:
        #
        #     self._simulate()
        # except Exception as e:
        #     Logger.message(e.__str__())
        #     QMessageBox.warning(self, 'Plotting Error', e.__str__(), QMessageBox.Ok)

    def fit(self):

        self._fit()
        # try:
        # except Exception as e:
        #     Logger.message(e.__str__())
        #     QMessageBox.warning(self, 'Fitting Error', e.__str__(), QMessageBox.Ok)

    def get_fit_range(self):
        try:
            x0_global, x1_global = float(self.leX0.text()), float(self.leX1.text())
        except ValueError:
            QMessageBox.warning(self, "Range", "Fitting range is not valid!", QMessageBox.Ok)
            return

        ranges = [[x0_global, x1_global] for _ in range(len(self.node))]
        model_bounds = self.pred_model_bounds if self.tabWidget.currentIndex() == 0 else self.general_model_bounds
        for lb, ub, rng in zip(model_bounds['lb'], model_bounds['ub'], ranges):
            x0_text, x1_text = lb.text(), ub.text()
            if x0_text.strip() == '' or x1_text.strip() == '':
                continue
            try:
                x0, x1 = float(x0_text), float(x1_text)
            except ValueError:
                continue
            rng[0] = x0
            rng[1] = x1

        return ranges

    def _simulate(self):
        model = self._predefined_model if self.tabWidget.currentIndex() == 0 else self._general_model

        if model is None:
            return

        ranges = self.get_fit_range()
        if ranges is None:
            return

        model.set_ranges(ranges)
        model.weight_func = self.res_weights[self.cbResWeighting.currentIndex()]['func']

        start_time = time.perf_counter()
        x_vals, fits, residuals = model.simulate()
        end_time = time.perf_counter()
        Logger.debug((end_time - start_time) * 1e3, 'ms for simulation')

        if model.varpro or (hasattr(model, 'fit_intercept_varpro') and model.fit_intercept_varpro):
            self.setup_fields()

        self.plot_fits(x_vals, fits, residuals)

    def _fit(self, run_non_blocking=True):

        if self.t_fit is not None and self.t_fit.isRunning():
            self.t_fit.requestInterruption()
            Logger.debug('Cancelling...')
            return

        model = self._predefined_model if self.tabWidget.currentIndex() == 0 else self._general_model

        if model is None:
            return

        ranges = self.get_fit_range()
        if ranges is None:
            return

        model.set_ranges(ranges)
        model.weight_func = self.res_weights[self.cbResWeighting.currentIndex()]['func']
        method = self.methods[self.cbMethod.currentIndex()]['abbr']

        self.t_fit = TaskFit(self, model, method, self.model_options_dict)
        if run_non_blocking:
            self.t_fit.start()
        else:  # runs with blocking of main thread
            self.t_fit.preRun()
            self.t_fit.run_fit()
            self.t_fit.postRun()

    @classmethod
    def replot(cls):
        if cls._instance is not None:
            cls._instance._replot()

    @classmethod
    def clear_plots(cls):
        if cls._instance is not None:
            cls._instance.clear_plot()

    def _replot(self):
        # if no fits simulated, return
        if len(self.fits) != len(self.node):
            return

        pw_plot_items = PlotWidget.plotted_items()
        df = Settings.fit_dark_factor

        for i, (sp_fit, sp_res) in enumerate(zip(self.fits, self.residuals)):
            fit_color = QColor(0, 0, 0, 255)
            res_color = QColor(0, 0, 200, 255)

            if self.node[i] in pw_plot_items:  # set fit_color of checked spectra as darkened
                plot_item = pw_plot_items[self.node[i]]
                node_color = plot_item.opts['pen'].color()
                if not self.cbFitBlackColor.isChecked():
                    fit_color.setRgbF(node_color.redF() * df,
                                      node_color.greenF() * df,
                                      node_color.blueF() * df,
                                      node_color.alphaF())

                    res_color.setRgbF(node_color.redF() * df * 0.9,
                                      node_color.greenF() * df * 0.9,
                                      node_color.blueF() * df * 0.9,
                                      node_color.alphaF())

                PlotWidget.plot_fit(sp_fit, name=f'Fif of {self.node[i].name}',
                                    pen=pg.mkPen(color=fit_color, width=2),
                                    zValue=2e5 - i)

                if self.cbShowResiduals.isChecked():
                    PlotWidget.plot_fit(sp_res, name=f'Residual of {self.node[i].name}',
                                        pen=pg.mkPen(color=res_color, width=1),
                                        zValue=1e5 - i)
                else:
                    PlotWidget.remove_fits([sp_res])
            else:
                # remove them from plot
                PlotWidget.remove_fits([sp_fit, sp_res])

    def plot_fits(self, x_vals, fits, residuals):

        assert len(fits) == len(residuals) == len(x_vals)

        for i in range(len(fits)):
            # only update data if spectra are already instantiated
            if len(self.fits) == len(fits):
                self.fits[i].data = np.vstack((x_vals[i], fits[i])).T
                self.residuals[i].data = np.vstack((x_vals[i], residuals[i])).T
            else:
                sp_fit = SpectrumItem.from_xy_values(x_vals[i], fits[i], name=f'Fif of {self.node[i].name}')
                sp_res = SpectrumItem.from_xy_values(x_vals[i], residuals[i], name=f'Residual of {self.node[i].name}')
                self.fits.children.append(sp_fit)
                self.residuals.children.append(sp_res)

        self._replot()

    # def _plot_function(self):
    #     x0, x1 = self.lr.getRegion()
    #
    #     # if this is a fit widget with real spectrum, use x range from that spectrum,
    #     # otherwise, use np.linspace with defined number of points in Settings
    #     if self.node is not None:
    #         start_idx = fi(self.node.data[:, 0], x0)
    #         end_idx = fi(self.node.data[:, 0], x1) + 1
    #         x_data = self.node.data[start_idx:end_idx, 0]
    #     else:
    #         x_data = np.linspace(x0, x1, num=Settings.FP_num_of_points)
    #
    #     tab_idx = self.tabWidget.currentIndex()
    #
    #     if tab_idx == 0:  # equation-based model
    #         self._setup_model()
    #
    #         par_count = self.current_model.par_count()
    #
    #         # get params from field to current model
    #         for i in range(par_count):
    #             param = self.params_list[i][0].text()
    #             self.current_model.params[param].value = float(self.value_list[i][0].text())
    #
    #         # calculate the y values according to our model
    #         y_data = self.current_model.wrapper_func(x_data, self.current_model.params)
    #     else:  # custom kinetic model
    #         if self.current_general_model is None:
    #             self.cbGenModel_changed()
    #
    #         init, coefs, rates, y0 = self.get_params_from_fields()
    #         par_count = 2 * init.shape[0] + rates.shape[0]
    #         n_params = len(init)
    #
    #         sol = self._simul_custom_model(init, rates, x_data)
    #
    #         if self.cbPlotAllComps.isChecked():
    #             spectra = SpectrumItemGroup(name='all comparments plot')
    #             for i in range(n_params):
    #                 name = self.params_list[i+n_params][tab_idx].text()
    #                 self.plotted_functions.append(self.plot_widget.plotItem.plot(x_data, sol[:, i],
    #                                                                          pen=pg.mkPen(color=int_default_color_scheme(i),
    #                                                                                       width=1),
    #                                                                          name=name))
    #                 spectra.children.append(SpectrumItem.from_xy_values(x_data, sol[:, i], name=name))
    #             self.plotted_function_spectra.append(spectra)
    #             return
    #
    #         y_data = (sol * coefs).sum(axis=1, keepdims=False) + y0
    #
    #     params = ', '.join([f"{self.params_list[i][tab_idx].text()}={self.value_list[i][tab_idx].text()}" for i in range(par_count)])
    #
    #     name = f"Func plot: {params}"
    #     self.plotted_functions.append(self.plot_widget.plotItem.plot(x_data, y_data,
    #                                                                  pen=pg.mkPen(color=QColor(0, 0, 0, 255),
    #                                                                               width=1),
    #                                                                  name=name))
    #
    #     self.plotted_function_spectra.append(SpectrumItem.from_xy_values(x_data, y_data, name=name))

    # def set_lower_upper_enabled(self):
    #     if self.node is None:
    #         return
    #     for i in range(len(self.fixed_list)):
    #         enabled_0 = not self.fixed_list[i][0].isChecked()
    #         enabled_1 = not self.fixed_list[i][1].isChecked()
    #
    #         self.lower_bound_list[i][0].setEnabled(enabled_0)
    #         self.upper_bound_list[i][0].setEnabled(enabled_0)
    #
    #         self.lower_bound_list[i][1].setEnabled(enabled_1)
    #         self.upper_bound_list[i][1].setEnabled(enabled_1)

    # def fixed_checked_changed(self, value):
    #     if self.node is None:
    #         return
    #
    #     checkbox = self.sender()  # get the checkbox that was clicked on
    #
    #     idx = self.tabWidget.currentIndex()
    #
    #     i = 0
    #     for i, (ch1, ch2) in enumerate(self.fixed_list):
    #         if ch1 == checkbox or ch2 == checkbox:
    #             break
    #
    #     enabled = True if value == Qt.Unchecked else False
    #
    #     self.lower_bound_list[i][idx].setEnabled(enabled)
    #     self.upper_bound_list[i][idx].setEnabled(enabled)

    # def params_count_changed(self):
    #
    #     count = int(self.sbParamsCount.value())
    #
    #     self.set_field_visible(count, 0)

    # def set_field_visible(self, n_params, idx=0):
    #     for i in range(self.max_param_count):
    #         visible = n_params > i
    #
    #         self.params_list[i][idx].setVisible(visible)
    #         self.lower_bound_list[i][idx].setVisible(visible)
    #         self.value_list[i][idx].setVisible(visible)
    #         self.upper_bound_list[i][idx].setVisible(visible)
    #         self.fixed_list[i][idx].setVisible(visible)
    #         self.error_list[i][idx].setVisible(visible)

    def set_result(self):
        pass

    def print_report(self):
        if self.fit_result is None:
            return

        try:
            report = f"Fitting of '{self.node[0].parent.name if len(self.node) > 0 else self.node[0].name}' succeeded. " \
                     f"You can find fit results in variable 'fit'." \
                     f"\n{self.fit_result.report(print=False)}\n"

            Logger.console_message(report)
            Console.push_variables({'fit': self.fit_result})

        except Exception as e:
            Logger.message(e.__str__())
            QMessageBox.warning(self, 'Printing Error', e.__str__(), QMessageBox.Ok)

    def accept(self):
        self.accepted_func()

        self.clear_plot()
        PlotWidget.remove_linear_region()

        self.print_report()

        self.accepted = True
        self.is_opened = False
        self._instance = None
        self.dock_widget.setVisible(False)

    def reject(self):
        self.clear_plot()
        PlotWidget.remove_linear_region()
        self.is_opened = False
        self._instance = None
        self.dock_widget.setVisible(False)


if __name__ == "__main__":
    import sys

    def my_exception_hook(exctype, value, traceback):
        # Print the error and traceback
        print(exctype, value, traceback)
        # Call the normal Exception hook after
        sys._excepthook(exctype, value, traceback)
        sys.exit(1)


    sys._excepthook = sys.excepthook

    # Set the exception hook to our wrapping function
    sys.excepthook = my_exception_hook

    app = QtWidgets.QApplication(sys.argv)
    Dialog = FitWidget(None, None)
    # Dialog.show()
    sys.exit(app.exec_())



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'spectramanipulator/dialogs\gui_fit.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(526, 664)
        self.gridLayout_2 = QtWidgets.QGridLayout(Dialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.verticalLayout.addWidget(self.label_2)
        self.horizontalLayout_3 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_3.setObjectName("horizontalLayout_3")
        self.cbModel = QtWidgets.QComboBox(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.cbModel.sizePolicy().hasHeightForWidth())
        self.cbModel.setSizePolicy(sizePolicy)
        self.cbModel.setObjectName("cbModel")
        self.horizontalLayout_3.addWidget(self.cbModel)
        self.btnPrintDiffEquations = QtWidgets.QPushButton(Dialog)
        self.btnPrintDiffEquations.setCheckable(False)
        self.btnPrintDiffEquations.setChecked(False)
        self.btnPrintDiffEquations.setAutoRepeat(False)
        self.btnPrintDiffEquations.setAutoDefault(False)
        self.btnPrintDiffEquations.setObjectName("btnPrintDiffEquations")
        self.horizontalLayout_3.addWidget(self.btnPrintDiffEquations)
        self.verticalLayout.addLayout(self.horizontalLayout_3)
        self.cbCustom = QtWidgets.QCheckBox(Dialog)
        self.cbCustom.setObjectName("cbCustom")
        self.verticalLayout.addWidget(self.cbCustom)
        self.label_13 = QtWidgets.QLabel(Dialog)
        self.label_13.setObjectName("label_13")
        self.verticalLayout.addWidget(self.label_13)
        self.cbMethod = QtWidgets.QComboBox(Dialog)
        self.cbMethod.setObjectName("cbMethod")
        self.verticalLayout.addWidget(self.cbMethod)
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setObjectName("label")
        self.verticalLayout.addWidget(self.label)
        self.pteEquation = QtWidgets.QPlainTextEdit(Dialog)
        sizePolicy = QtWidgets.QSizePolicy(QtWidgets.QSizePolicy.Preferred, QtWidgets.QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.pteEquation.sizePolicy().hasHeightForWidth())
        self.pteEquation.setSizePolicy(sizePolicy)
        self.pteEquation.setMinimumSize(QtCore.QSize(0, 200))
        self.pteEquation.setMaximumSize(QtCore.QSize(16777215, 200))
        self.pteEquation.setPlainText("")
        self.pteEquation.setObjectName("pteEquation")
        self.verticalLayout.addWidget(self.pteEquation)
        self.label_10 = QtWidgets.QLabel(Dialog)
        self.label_10.setObjectName("label_10")
        self.verticalLayout.addWidget(self.label_10)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.label_11 = QtWidgets.QLabel(Dialog)
        self.label_11.setObjectName("label_11")
        self.horizontalLayout.addWidget(self.label_11)
        self.leX0 = QtWidgets.QLineEdit(Dialog)
        self.leX0.setObjectName("leX0")
        self.horizontalLayout.addWidget(self.leX0)
        self.label_12 = QtWidgets.QLabel(Dialog)
        self.label_12.setObjectName("label_12")
        self.horizontalLayout.addWidget(self.label_12)
        self.leX1 = QtWidgets.QLineEdit(Dialog)
        self.leX1.setObjectName("leX1")
        self.horizontalLayout.addWidget(self.leX1)
        self.label_9 = QtWidgets.QLabel(Dialog)
        self.label_9.setObjectName("label_9")
        self.horizontalLayout.addWidget(self.label_9)
        self.sbParamsCount = QtWidgets.QSpinBox(Dialog)
        self.sbParamsCount.setMinimum(1)
        self.sbParamsCount.setMaximum(10)
        self.sbParamsCount.setObjectName("sbParamsCount")
        self.horizontalLayout.addWidget(self.sbParamsCount)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.cbUpdateInitValues = QtWidgets.QCheckBox(Dialog)
        self.cbUpdateInitValues.setObjectName("cbUpdateInitValues")
        self.verticalLayout.addWidget(self.cbUpdateInitValues)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.label_7 = QtWidgets.QLabel(Dialog)
        self.label_7.setAlignment(QtCore.Qt.AlignCenter)
        self.label_7.setObjectName("label_7")
        self.gridLayout.addWidget(self.label_7, 0, 4, 1, 1)
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setAlignment(QtCore.Qt.AlignCenter)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 0, 0, 1, 1)
        self.label_8 = QtWidgets.QLabel(Dialog)
        self.label_8.setAlignment(QtCore.Qt.AlignCenter)
        self.label_8.setObjectName("label_8")
        self.gridLayout.addWidget(self.label_8, 0, 5, 1, 1)
        self.label_6 = QtWidgets.QLabel(Dialog)
        self.label_6.setAlignment(QtCore.Qt.AlignCenter)
        self.label_6.setObjectName("label_6")
        self.gridLayout.addWidget(self.label_6, 0, 3, 1, 1)
        self.label_4 = QtWidgets.QLabel(Dialog)
        self.label_4.setAlignment(QtCore.Qt.AlignCenter)
        self.label_4.setObjectName("label_4")
        self.gridLayout.addWidget(self.label_4, 0, 1, 1, 1)
        self.label_5 = QtWidgets.QLabel(Dialog)
        self.label_5.setAlignment(QtCore.Qt.AlignCenter)
        self.label_5.setObjectName("label_5")
        self.gridLayout.addWidget(self.label_5, 0, 2, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.verticalLayout.addItem(spacerItem)
        self.horizontalLayout_2 = QtWidgets.QHBoxLayout()
        self.horizontalLayout_2.setObjectName("horizontalLayout_2")
        self.btnFit = QtWidgets.QPushButton(Dialog)
        self.btnFit.setCheckable(False)
        self.btnFit.setChecked(False)
        self.btnFit.setAutoRepeat(False)
        self.btnFit.setAutoDefault(False)
        self.btnFit.setObjectName("btnFit")
        self.horizontalLayout_2.addWidget(self.btnFit)
        self.btnPrintReport = QtWidgets.QPushButton(Dialog)
        self.btnPrintReport.setCheckable(False)
        self.btnPrintReport.setChecked(False)
        self.btnPrintReport.setAutoRepeat(False)
        self.btnPrintReport.setAutoDefault(False)
        self.btnPrintReport.setObjectName("btnPrintReport")
        self.horizontalLayout_2.addWidget(self.btnPrintReport)
        spacerItem1 = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.horizontalLayout_2.addItem(spacerItem1)
        self.btnOK = QtWidgets.QPushButton(Dialog)
        self.btnOK.setAutoDefault(False)
        self.btnOK.setObjectName("btnOK")
        self.horizontalLayout_2.addWidget(self.btnOK)
        self.btnCancel = QtWidgets.QPushButton(Dialog)
        self.btnCancel.setAutoDefault(False)
        self.btnCancel.setObjectName("btnCancel")
        self.horizontalLayout_2.addWidget(self.btnCancel)
        self.verticalLayout.addLayout(self.horizontalLayout_2)
        self.gridLayout_2.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(Dialog)
        QtCore.QMetaObject.connectSlotsByName(Dialog)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.label_2.setText(_translate("Dialog", "Model:"))
        self.btnPrintDiffEquations.setText(_translate("Dialog", "Print differential equations"))
        self.cbCustom.setText(_translate("Dialog", "Custom model (equation window will be editable)"))
        self.label_13.setText(_translate("Dialog", "Minimizing algorithm:"))
        self.label.setText(_translate("Dialog", "<html><head/><body><p><span style=\" font-weight:600;\">def func(x, *params):</span></p></body></html>"))
        self.label_10.setText(_translate("Dialog", "Selected x range:"))
        self.label_11.setText(_translate("Dialog", "from"))
        self.label_12.setText(_translate("Dialog", "to"))
        self.label_9.setText(_translate("Dialog", "Params"))
        self.cbUpdateInitValues.setText(_translate("Dialog", "Update initial values when linear region move is finished"))
        self.label_7.setText(_translate("Dialog", "Fixed"))
        self.label_3.setText(_translate("Dialog", "Param"))
        self.label_8.setText(_translate("Dialog", "Error"))
        self.label_6.setText(_translate("Dialog", "Upper\n"
"bound"))
        self.label_4.setText(_translate("Dialog", "Lower\n"
"bound"))
        self.label_5.setText(_translate("Dialog", "<= Value <="))
        self.btnFit.setText(_translate("Dialog", "Fit"))
        self.btnPrintReport.setText(_translate("Dialog", "Print report"))
        self.btnOK.setText(_translate("Dialog", "Ok"))
        self.btnCancel.setText(_translate("Dialog", "Cancel"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())



import xml.etree.ElementTree as ET
from spectramanipulator.spectrum import Spectrum
import numpy as np

# bizare stuff...
def parse_XML_Spreadsheet(xml_text):
    def float_try_parse(num):
        try:
            return float(num.replace(',', '.').strip())
        except ValueError:
            return None

    try:
        root = ET.fromstring(xml_text)
        ssNs = '{urn:schemas-microsoft-com:office:spreadsheet}'

        # Table element
        t_el = None

        for el in root.iter(ssNs + 'Table'):
            t_el = el

        col_count = int(t_el.get(ssNs + 'ExpandedColumnCount'))

        if col_count < 2:
            return

        errs = col_count * [False]
        names = col_count * ['']

        buffer = []

        for row in t_el.iter(ssNs + 'Row'):

            vals = col_count * [None]

            # check whether current row contains only number, all of the Data elements are Type="Number"
            contains_numbers = all(el.attrib[ssNs + 'Type'] == 'Number' for el in row.iter(ssNs + 'Data'))

            cells = row.iter(ssNs + 'Cell')
            cell_idx = 0

            for cell in cells:

                index_attr = None
                try:
                    index_attr = cell.attrib[ssNs + "Index"]
                except KeyError:
                    pass

                if index_attr is not None:
                    new_cell_idx = int(index_attr) - 1
                    if contains_numbers:
                        for i in range(cell_idx, new_cell_idx):
                            errs[i] = True
                    else:
                        pass

                    cell_idx = new_cell_idx

                query = [el for el in cell.iter(ssNs + 'Data')]

                if len(query) == 0:
                    continue

                el_data = query[0]

                if contains_numbers:
                    vals[cell_idx] = float_try_parse(el_data.text)
                    if vals[cell_idx] is None:
                        errs[cell_idx] = True
                else:
                    names[cell_idx] = str(el_data.text)

                cell_idx += 1

            if cell_idx == 0:  # no cells in current row
                continue
            if cell_idx != col_count:
                for i in range(cell_idx, col_count):
                    errs[i] = True

            if contains_numbers:
                buffer.append(vals)

        row_count = len(buffer)

        if row_count < 2:
            return

        first_occur_idx = -1
        for i in range(col_count):
            if not errs[i]:
                first_occur_idx = i
                break

        if first_occur_idx == -1:
            return

        spectra = []

        for i in range(first_occur_idx + 1, col_count):
            # skip the wrong column
            if errs[i]:
                continue

            sp_data = []
            # group_name = self.get_filename() if self.filepath is not None else ''

            for j in range(row_count):
                sp_data.append((buffer[j][first_occur_idx], buffer[j][i]))

            sp = Spectrum(np.asarray(sp_data, dtype=np.float64), name=names[i])

            spectra.append(sp)

        return spectra if len(spectra) < 2 else [spectra]
    except Exception:
        raise

# if __name__ == "__main__":
#
#
#
# xml_data = """<?xml version="1.0"?>
# <?mso-application progid="Excel.Sheet"?>
# <Workbook xmlns="urn:schemas-microsoft-com:office:spreadsheet"
#  xmlns:o="urn:schemas-microsoft-com:office:office"
#  xmlns:x="urn:schemas-microsoft-com:office:excel"
#  xmlns:ss="urn:schemas-microsoft-com:office:spreadsheet"
#  xmlns:html="http://www.w3.org/TR/REC-html40">
#  <Styles>
#   <Style ss:ID="Default" ss:Name="Normal">
#    <Alignment ss:Vertical="Bottom"/>
#    <Borders/>
#    <Font ss:FontName="Calibri" x:CharSet="238" x:Family="Swiss" ss:Size="11"
#     ss:Color="#000000"/>
#    <Interior/>
#    <NumberFormat/>
#    <Protection/>
#   </Style>
#   <Style ss:ID="s66">
#    <NumberFormat ss:Format="0.0"/>
#   </Style>
#  </Styles>
#  <Worksheet ss:Name="Sheet1">
#   <Table ss:ExpandedColumnCount="4" ss:ExpandedRowCount="4"
#    ss:DefaultRowHeight="14.4">
#    <Row>
#     <Cell><Data ss:Type="String">wl</Data></Cell>
#     <Cell><Data ss:Type="String">a</Data></Cell>
#     <Cell><Data ss:Type="String">b</Data></Cell>
#     <Cell><Data ss:Type="String">c</Data></Cell>
#    </Row>
#    <Row>
#     <Cell ss:StyleID="s66"><Data ss:Type="Number">1.48613546846513</Data></Cell>
#     <Cell><Data ss:Type="Number">1567</Data></Cell>
#     <Cell><Data ss:Type="Number">11</Data></Cell>
#     <Cell><Data ss:Type="Number">266</Data></Cell>
#    </Row>
#    <Row>
#     <Cell><Data ss:Type="Number">5.87</Data></Cell>
#     <Cell><Data ss:Type="Number">458</Data></Cell>
#     <Cell><Data ss:Type="Number">48</Data></Cell>
#     <Cell><Data ss:Type="Number">55</Data></Cell>
#    </Row>
#    <Row>
#     <Cell><Data ss:Type="Number">67</Data></Cell>
#     <Cell><Data ss:Type="Number">4852</Data></Cell>
#     <Cell><Data ss:Type="Number">996</Data></Cell>
#     <Cell><Data ss:Type="Number">11.56</Data></Cell>
#    </Row>
#   </Table>
#  </Worksheet>
# </Workbook>"""
#
#
#
#
# # ee = ET.Element()
#
#
# def ns(element):
#     return '{urn:schemas-microsoft-com:office:spreadsheet}' + element
#
# def float_try_parse(num):
#     try:
#         return float(num.replace('.', '.').strip())
#     except ValueError:
#         return None
#
# root = ET.fromstring(xml_data)
# ssNs = '{urn:schemas-microsoft-com:office:spreadsheet}'
#
# # Table element
# t_el = None
#
# for el in root.iter(ssNs + 'Table'):
#     t_el = el
#
# col_count = int(t_el.get(ssNs + 'ExpandedColumnCount'))
#
# if col_count < 2:
#     print("return")
#     exit(0)
#
# errs = col_count * [False]
# names = col_count * ['']
#
# buffer = []
#
#
# for row in t_el.iter(ssNs + 'Row'):
#
#     vals = col_count * [None]
#
#     # check whether current row contains only number, all of the Data elements are Type="Number"
#     contains_numbers = all(el.attrib[ssNs + 'Type'] == 'Number' for el in row.iter(ssNs + 'Data'))
#
#     cells = row.iter(ssNs + 'Cell')
#     cell_idx = 0
#
#     for cell in cells:
#
#         index_attr = None
#         try:
#             index_attr = cell.attrib[ssNs + "Index"]
#         except KeyError:
#             pass
#
#         if index_attr is not None:
#             new_cell_idx = int(index_attr) - 1
#             if contains_numbers:
#                 for i in range(cell_idx, new_cell_idx):
#                     errs[i] = True
#             else:
#                 pass
#
#             cell_idx = new_cell_idx
#
#         query = [el for el in cell.iter(ssNs + 'Data')]
#
#         if len(query) == 0:
#             continue
#
#         el_data = query[0]
#
#         if contains_numbers:
#             vals[cell_idx] = float_try_parse(el_data.text)
#             if vals[cell_idx] is None:
#                 errs[cell_idx] = True
#         else:
#             names[cell_idx] = str(el_data.text)
#
#         cell_idx += 1
#
#     if cell_idx == 0:  # no cells in current row
#         continue
#     if cell_idx != col_count:
#         for i in range(cell_idx, col_count):
#             errs[i] = True
#
#     if contains_numbers:
#         buffer.append(vals)
#
# row_count = len(buffer)
#
# if row_count < 2:
#     print("return")
#     exit(0)
#
# first_occur_idx = -1
# for i in range(col_count):
#     if not errs[i]:
#         first_occur_idx = i
#         break
#
# if first_occur_idx == -1:
#     print("return")
#     exit(0)
#
# spectra = []
#
# for i in range(first_occur_idx + 1, col_count):
#     # skip the wrong column
#     if errs[i]:
#         continue
#
#     sp_data = []
#     # group_name = self.get_filename() if self.filepath is not None else ''
#
#     for j in range(row_count):
#         sp_data.append((buffer[j][first_occur_idx], buffer[j][i]))
#
#     sp = Spectrum(np.asarray(sp_data, dtype=np.float64), name=names[i])
#
#     spectra.append(sp)
#
#
#
#
#
#



import sys


class Logger(object):

    _instance = None

    def __init__(self, console_message, statusbar_message):
        Logger._instance = self

        # functions
        self.statusbar_message = statusbar_message
        self.console_message = console_message

    @classmethod
    def console_message(cls, text, add_new_line=True):
        if cls._instance is None:
            return

        cls._instance.console_message(str(text), add_new_line)

    @classmethod
    def status_message(cls, text, delay=3000):
        if cls._instance is None:
            return

        cls._instance.statusbar_message(str(text), delay)

    @classmethod
    def message(cls, text, delay=3000):
        if cls._instance is None:
            return

        cls._instance.console_message(str(text))
        cls._instance.statusbar_message(str(text), delay)

    @staticmethod
    def debug(*values):
        from spectramanipulator.__main__ import debug
        if debug:
            print(*values)


class Transcript:
    """Class used to redirect std output to pyqtconsole"""

    def __init__(self):
        self.terminal = sys.stdout

    def __getattr__(self, attr):
        return getattr(self.terminal, attr)

    def write(self, message):
        self.terminal.write(message)
        Logger.console_message(message, False)

    def flush(self):
        pass




# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'spectramanipulator/dialogs\gui_export_spectra_as.ui'
#
# Created by: PyQt5 UI code generator 5.15.2
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Dialog(object):
    def setupUi(self, Dialog):
        Dialog.setObjectName("Dialog")
        Dialog.resize(416, 266)
        self.gridLayout_2 = QtWidgets.QGridLayout(Dialog)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.verticalLayout = QtWidgets.QVBoxLayout()
        self.verticalLayout.setObjectName("verticalLayout")
        self.lblTitle = QtWidgets.QLabel(Dialog)
        self.lblTitle.setWordWrap(True)
        self.lblTitle.setObjectName("lblTitle")
        self.verticalLayout.addWidget(self.lblTitle)
        self.horizontalLayout = QtWidgets.QHBoxLayout()
        self.horizontalLayout.setObjectName("horizontalLayout")
        self.leDir = QtWidgets.QLineEdit(Dialog)
        self.leDir.setObjectName("leDir")
        self.horizontalLayout.addWidget(self.leDir)
        self.btnDir = QtWidgets.QToolButton(Dialog)
        self.btnDir.setObjectName("btnDir")
        self.horizontalLayout.addWidget(self.btnDir)
        self.verticalLayout.addLayout(self.horizontalLayout)
        self.gridLayout = QtWidgets.QGridLayout()
        self.gridLayout.setObjectName("gridLayout")
        self.leFileExt = QtWidgets.QLineEdit(Dialog)
        self.leFileExt.setObjectName("leFileExt")
        self.gridLayout.addWidget(self.leFileExt, 0, 2, 1, 1)
        self.leDelimiter = QtWidgets.QLineEdit(Dialog)
        self.leDelimiter.setObjectName("leDelimiter")
        self.gridLayout.addWidget(self.leDelimiter, 1, 2, 1, 1)
        self.label_2 = QtWidgets.QLabel(Dialog)
        self.label_2.setObjectName("label_2")
        self.gridLayout.addWidget(self.label_2, 1, 1, 1, 1)
        self.leDecimalSeparator = QtWidgets.QLineEdit(Dialog)
        self.leDecimalSeparator.setObjectName("leDecimalSeparator")
        self.gridLayout.addWidget(self.leDecimalSeparator, 3, 2, 1, 1)
        self.label = QtWidgets.QLabel(Dialog)
        self.label.setObjectName("label")
        self.gridLayout.addWidget(self.label, 0, 1, 1, 1)
        self.label_3 = QtWidgets.QLabel(Dialog)
        self.label_3.setObjectName("label_3")
        self.gridLayout.addWidget(self.label_3, 3, 1, 1, 1)
        spacerItem = QtWidgets.QSpacerItem(40, 20, QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Minimum)
        self.gridLayout.addItem(spacerItem, 0, 3, 1, 1)
        self.verticalLayout.addLayout(self.gridLayout)
        self.buttonBox = QtWidgets.QDialogButtonBox(Dialog)
        self.buttonBox.setOrientation(QtCore.Qt.Horizontal)
        self.buttonBox.setStandardButtons(QtWidgets.QDialogButtonBox.Cancel|QtWidgets.QDialogButtonBox.Save)
        self.buttonBox.setObjectName("buttonBox")
        self.verticalLayout.addWidget(self.buttonBox)
        self.gridLayout_2.addLayout(self.verticalLayout, 0, 0, 1, 1)

        self.retranslateUi(Dialog)
        self.buttonBox.accepted.connect(Dialog.accept)
        self.buttonBox.rejected.connect(Dialog.reject)
        QtCore.QMetaObject.connectSlotsByName(Dialog)
        Dialog.setTabOrder(self.btnDir, self.leDir)

    def retranslateUi(self, Dialog):
        _translate = QtCore.QCoreApplication.translate
        Dialog.setWindowTitle(_translate("Dialog", "Dialog"))
        self.lblTitle.setText(_translate("Dialog", "Each selected top-level item will be saved in separate file with its name as filename. Spectra in groups will be concatenated. Select directory where spectra will be saved and specify the file extension. Non existing directories will be created. Top level items with same name will be overwritten. Delimiter field: use \\t for tabulator."))
        self.btnDir.setText(_translate("Dialog", "..."))
        self.label_2.setText(_translate("Dialog", "Delimiter:"))
        self.label.setText(_translate("Dialog", "File extension:"))
        self.label_3.setText(_translate("Dialog", "Decimal separator:"))


if __name__ == "__main__":
    import sys
    app = QtWidgets.QApplication(sys.argv)
    Dialog = QtWidgets.QDialog()
    ui = Ui_Dialog()
    ui.setupUi(Dialog)
    Dialog.show()
    sys.exit(app.exec_())




import numpy as np
# import math
from abc import abstractmethod

from scipy.signal import savgol_filter
from scipy.interpolate import interp1d
from scipy.fftpack import fft, fftfreq
from scipy.integrate import cumtrapz, simps

import functools


## from https://github.com/Tillsten/skultrafast/blob/master/skultrafast/dv.py
def fi(array, values):
    """
    Finds index of nearest `value` in `array`. If value >  max(array), the last index of array
    is returned, if value < min(array), 0 is returned. Array must be sorted. Also works for value
    to be array, then array of indexes is returned.

    Parameters
    ----------
    array : ndarray
        Array to be searched.
    values : {int, float, list}
        Value or values to look for.

    Returns
    -------
    out : int, np.ndarray
        Found nearest index/es to value/s.
    """
    try:
        len(values)
    except TypeError:
        values = [values]
    ret_idx = [np.argmin(np.abs(array - i)) for i in values]

    return ret_idx[0] if len(ret_idx) == 1 else ret_idx


# def find_nearest(array, value):
#     """
#     Returns closest value in in `array`. If value >  max(array), the last value of array
#     is returned, if value < min(array), first value of array is returned. Array must be sorted.
#
#     Parameters
#     ----------
#     array : ndarray
#         Array to be searched.
#     value : {int, float}
#         Value.
#
#     Returns
#     -------
#     out : type of values in `array`
#         Found nearest value.
#     """
#     idx = fi(array, value)
#     return array[idx]


def group2mat(spectra):
    """Given a generator of spectra (SpectrumList, list, etc.), it returns the x values, 'name'
     values converted to float and value matrix in shape of [x, y]. All spectra has to have in
     same shape. If some names cannot be converted to float, None is returned for y values.

    Parameters
    ----------
    spectra : iterable
        Generator of spectra (SpectrumList, list, etc.)

    Returns
    -------
    out : tuple
        x_values, converted_names as y values, matrix with [x, y] dimension"""

    header_vals_temp = []

    x = None
    matrix = None
    err = False

    for sp in spectra:
        try:
            header_vals_temp.append(float(sp.name.replace(',', '.').strip()))
        except ValueError:
            err = True

        if x is None:
            x = sp.data[:, 0]
        else:
            if x.shape[0] != sp.data.shape[0]:
                raise ValueError("Spectra do not have the same dimension within the group. Unable to perform the operation.")
            if not np.allclose(x, sp.data[:, 0]):
                raise ValueError("Spectra do not have the same x values within the group. Unable to perform the operation.")

        matrix = sp.data[:, 1] if matrix is None else np.vstack((matrix, sp.data[:, 1]))

    y = None if err else np.asarray(header_vals_temp)

    return x, y, matrix.T


def operation(operator=lambda a, b: a + b, operator_str='+', switch_names=False):
    """TODO--->"""
    def decorator(fn):
        @functools.wraps(fn)
        def func(self, other):
            obj, other_str = self._arithmetic_operation(other, operator)
            if obj is None:
                return NotImplemented
            if switch_names:
                obj.name = f"{other_str} {operator_str} {self.name}"
            else:
                obj.name = f"{self.name} {operator_str} {other_str}"

            return obj
        return func
    return decorator


class IOperationBase:
    """
    Interface that defines the operations for Spectrum and SpectrumList objects.
    It allows for ... TODO...
    """

    def __init__(self, name):
        self.name = name

    def _redraw_all_spectra(self):
        pass

    def _update_view(self):
        pass

    @abstractmethod
    def _arithmetic_operation(self, other, operator):
        """TODO--->>"""
        pass

    @operation(lambda a, b: a + b, '+', False)  # self + other
    def __add__(self, other):
        pass

    @operation(lambda a, b: a - b, '-', False)  # self - other
    def __sub__(self, other):
        pass

    @operation(lambda a, b: a * b, '*', False)  # self * other
    def __mul__(self, other):
        pass

    @operation(lambda a, b: a / b, '/', False)  # self / other
    def __truediv__(self, other):
        pass

    @operation(lambda a, b: b + a, '+', True)  # other + self, switch_names=True
    def __radd__(self, other):
        pass

    @operation(lambda a, b: b - a, '-', True)  # other - self, switch_names=True
    def __rsub__(self, other):
        pass

    @operation(lambda a, b: b * a, '*', True)  # other * self, switch_names=True
    def __rmul__(self, other):
        pass

    @operation(lambda a, b: b / a, '/', True)  # other / self, switch_names=True
    def __rtruediv__(self, other):
        pass

    @operation(lambda a, b: a ** b, '**', False)  # self ** power
    def __pow__(self, power, modulo=None):
        pass

    @operation(lambda a, b: b ** a, '**', True)  # other ** self, switch_names=True
    def __rpow__(self, other):
        pass

    def __neg__(self):  # - self
        obj = self * -1
        obj.name = f'-{self.name}'
        return obj

    def __pos__(self):  # + self
        return self.__copy__()

    def __copy__(self):
        return NotImplementedError


class SpectrumList(IOperationBase):

    def __init__(self, children=None, name=''):
        """
        Class that holds a group of spectra as a list and enables simple arithmetic calculations among groups of spectra, number and
        group and spectrum and group. It acts as a list, so the individual spectra can be accessed by putting square brackets
        after the instance of an object and putting appropriate index in them. Also, spectrum objects can be iterated the same way as list.
        Also, it enables other calculations on the group of spectra, like baseline correction, cutting, basically all the
        operation provided in :class:`Spectrum` class plus additional operations that can
        be only used for groups, like transposition, etc. ....

        Parameters
        ----------
        children : {list, None}
            List of :class:`Spectrum`, default None.
        name : str, optional
            The group name.

        Attributes
        ----------
        children : list of :class:`Spectrum`
            List of members in this group.
        name : str
            The group name as appeared in Tree Widget.
        """

        super(SpectrumList, self).__init__(name)
        self.children = [] if children is None else children

    def __len__(self):
        return self.children.__len__()

    def get_names(self):
        """Returns names of all Spectrum objects as a list.

        :returns: list
        """

        return [sp.name for sp in self]

    # update all spectra, view
    def set_names(self, names):
        """Sets the names of the group and updates the Tree Widget and redraws the spectra.

        Parameters
        ----------
        names : iterable, eg. list, tuple
            New names that will be replaced. The length of names can be different from number
            of spectra in group (using a zip function).
        """

        for sp, new_name in zip(self, names):
            sp.name = str(new_name)

    def get_y_values_at_x(self, x):
        """Returns y values at particular x value as an ndarray in this group of spectra.

        Parameters
        ----------
        x : {int, float}
            The x value.

        Returns
        -------
        out : ndarray
            Array of y values at this x value.
        """

        ret_list = []

        for sp in self:
            idx = fi(sp.data[:, 0], x)
            ret_list.append(sp.data[idx, 1])

        return np.asarray(ret_list, dtype=np.float64)

    def get_transpose(self, max_items=1000):
        """
        Transposes group, names of spectra in group will be taken as x values for transposed data, so these values
        must be convertible to int or float numbers. No text is allowed in these cells, only values.
        All x values of spectra in the group will become names in new group. The transposed group is added to Tree Widget.
        The operation is basically the same as copying the group of spectra to Excel, performs transposition of the matrix
        and copying back to spectramanipulator.

        Parameters
        ----------
        max_items : int
            Maximum number of items that will be added. If the transposition produce more than *max_items*,
            ValueError will be raised. Default 1000.
        """

        if len(self) < 2:
            raise ValueError("At least 2 items have to be in the group in order to perform transposition.")

        if not isinstance(self[0], Spectrum):
            raise ValueError("Objects in list have to be type of Spectrum.")

        x, y, matrix = group2mat(self.__iter__())

        if y is None:
            raise ValueError("Names of spectra cannot be parsed to float.")

        n = x.shape[0]
        assert n == matrix.shape[0]

        group_name = "Transpose of {}".format(self.name)
        spectra = []

        if n > max_items:
            raise ValueError(
                "Number of transposed items ({}) exceeded the maximum number ({}). Cannot transpose spectra.".format(
                    n, max_items))

        for i in range(n):
            sp = Spectrum.from_xy_values(y, matrix[i], name=str(x[i]))
            spectra.append(sp)

        return SpectrumList(spectra, name=group_name)

    def _arithmetic_operation(self, other, func_operation):
        """
        Perform an defined operation on groups of spectra (type Spectrum). func_operation is a pointer to function.
        This function takes 2 arguments that must be type of Spectrum.
        """
        # operation with another list, group + group
        if isinstance(other, (SpectrumList, np.ndarray)):
            if len(self) != len(other):
                raise ValueError("Cannot perform an operation on groups which contains different number of items.")
            if len(self) == 0:
                return SpectrumList()
            # if not isinstance(self[0], Spectrum) or not isinstance(other[0], Spectrum):
            #     raise ValueError("Objects in list have to be type of Spectrum.")
            ret_list = SpectrumList()
            for i in range(len(self)):
                ret_list.children.append(func_operation(self[i], other[i]))
            return ret_list, other.name if isinstance(other, SpectrumList) else 'ndarray'

        # operation with single spectrum, group + spectrum or with number, eg. group - 1
        elif isinstance(other, (Spectrum, float, int)):
            if len(self) == 0:
                return SpectrumList()

            if not isinstance(self[0], Spectrum):
                raise ValueError("Objects in list have to be type of Spectrum.")
            ret_list = SpectrumList()
            for sp in self:
                ret_list.children.append(func_operation(sp, other))
            return ret_list, str(other) if isinstance(other, (float, int)) else other.name

        return None, None

    def __neg__(self):
        sl = SpectrumList()
        sl.name = f'-{self.name}'
        for sp in self:
            sl.children.append(-sp)
        return sl

    def __getitem__(self, item):
        return self.children[item]

    def __iter__(self):
        return iter(self.children)

    def __repr__(self):
        if self.__len__() == 0:  # self.__class__.__name__ gets the name of the current class instance
            return f'{self.__class__.__name__} without data'
        return f"{self.__class__.__name__}({self.name}, {self.__len__()} items)"

    def __copy__(self):
        """Deep copy this instance as SpectrumList."""
        ret = SpectrumList(name=self.name)
        for child in self.children:
            ret.children.append(child.__copy__())
        return ret


def add_modif_func(redraw_spectra=True, update_view=False):
    """Adds the function to cls
    adds a new functions with """
    # https://stackoverflow.com/questions/2366713/can-a-decorator-of-an-instance-method-access-the-class

    class Decorator:
        def __init__(self, fn):
            self.fn = fn

        def __set_name__(self, owner, fn_name):
            # gets called when the owner class is created !!

            # do something with owner, i.e.
            # print(f"decorating {self.fn} and using {owner}")
            # self.fn.class_name = owner.__name__

            @functools.wraps(self.fn)
            def fn_spectrum(this, *args, **kwargs):
                self.fn(this, *args, **kwargs)

                if redraw_spectra:
                    this._redraw_all_spectra()
                if update_view:
                    this._update_view()

                return this

            @functools.wraps(self.fn)
            def fn_spectrum_list_no_update(this, *args, **kwargs):
                for sp in this:
                    self.fn(sp, *args, **kwargs)  # perform the operation for each spectrum

                return this

            @functools.wraps(self.fn)
            def fn_spectrum_list(this, *args, **kwargs):
                for sp in this:
                    self.fn(sp, *args, **kwargs)  # perform the operation for each spectrum

                if redraw_spectra:
                    this._redraw_all_spectra()
                if update_view:
                    this._update_view()

                return this

            # then replace ourself with the original method
            setattr(owner, f'{fn_name}_no_update', self.fn)  # no update as with original function
            setattr(owner, fn_name, fn_spectrum)  # use same name for modified fcn
            setattr(SpectrumList, f'{fn_name}_no_update', fn_spectrum_list_no_update)
            setattr(SpectrumList, fn_name, fn_spectrum_list)

    return Decorator


def add_op_func():

    def decorator(fn):
        @functools.wraps(fn)
        def fn_spectrum_list(self, *args, **kwargs):
            # perform the operation for each spectrum and collect the results as ndarray
            return np.asarray([fn(sp, *args, **kwargs) for sp in self])

        setattr(SpectrumList, fn.__name__, fn_spectrum_list)

        return fn

    return decorator


class Spectrum(IOperationBase):

    def __init__(self, data=None, name='', filepath=None, assume_sorted=False, **kwargs):
        """
        Class that holds the spectrum object as a 2D array (dimensions n x 2) where n is a number
        of points in spectrum and includes various functions used for data manipulation with them.
        Spectrum is stored in variable *data* as a numpy ndarray.

        Parameters
        ----------
        data : numpy.ndarray
            A raw data of spectrum object, dimensions are (n x 2) where n is a number of points in spectrum.
            First column represents *x* values, second column *y* values.
        name : str
            The name of the spectrum as is displayed in Tree Widget.
        filepath : {str, None}
            If the spectrum was imported from file, this variable stored the path to the file it was imported from.
        """

        super(Spectrum, self).__init__(name)

        if assume_sorted:
            self.data = np.asarray(data, dtype=np.float64) if data is not None else None
        else:
            # sort according to first column,numpy matrix, 1. column wavelength, 2. column absorbance
            self.data = np.asarray(data[data[:, 0].argsort()], dtype=np.float64) if data is not None else None

        self.filepath = filepath

    @classmethod
    def from_xy_values(cls, x_values, y_values, name='', filepath=None, **kwargs):
        """
        Creates the Spectrum object from separate x and y data variables. The dimensions of x_values
        and y_values must be the same and contain numbers. The other parameters are the same
        as for __init__ method.

        Parameters
        ----------
        x_values : iterable, eg. list, tuple, ndarray
            Iterable that represents x values.
        y_values : iterable, eg. list, tuple, ndarray
            Iterable that represents y values.

        Returns
        -------
        out : :class:`Spectrum`
            Spectrum object.


        Raises
        ------
        ValueError
            If *x_values* or *y_values* have not the same dimension or do not contain numbers.
        """

        # check whether the x and y values can be iterated over
        if not np.iterable(x_values) and not np.iterable(y_values):
            raise ValueError(
                "Argument error, x_values and y_values must be iterables and contain numbers.")

        try:
            if len(x_values) != len(y_values):
                raise ValueError("Length of x_values and y_values must match.")

            x_data = np.asarray(x_values, dtype=np.float64)
            y_data = np.asarray(y_values, dtype=np.float64)
        except ValueError:
            raise

        data = np.vstack((x_data, y_data)).T

        return cls(data=data, name=name, filepath=filepath, assume_sorted=False, **kwargs)

    @property
    def x(self):
        """Returns the x values of this spectrum.

        :return: ndarray
        """
        return self.data[:, 0]

    @x.setter
    def x(self, array):
        """Sets the x values of this spectrum.

        :return: ndarray
        """
        self.data[:, 0] = array

    @property
    def y(self):
        """Returns the y values of this spectrum.

        :return: ndarray
        """
        return self.data[:, 1]

    @y.setter
    def y(self, array):
        """Sets the y values of this spectrum.

        :return: ndarray
        """
        self.data[:, 1] = array

    def length(self):
        """Returns the number of point that this spectrum has.

        :return: float
        """
        return self.data.shape[0]

    @add_modif_func(True, True)
    def power_spectrum(self):
        # https://stackoverflow.com/questions/15382076/plotting-power-spectrum-in-python
        """Calculates the power spectrum using FFT and replaces the data values. Power spectrum is normalized to 1.
        """

        n = self.data.shape[0]  # number of points in current spectrum

        vals = np.abs(fft(self.data[:, 1])) ** 2
        vals /= np.max(vals)
        freq = fftfreq(n, (self.data[-1, 0] - self.data[0, 0]) / (n - 1))

        filter = freq >= 0  # pick only frequencies >= 0

        output = np.vstack((freq[filter], vals[filter])).T

        self.data = output

        return self

    @add_modif_func(True, False)
    def savgol(self, window_length, poly_order):
        """
        Applies Savitsky-Golay filter to a spectrum. Based on `scipy.signal.savgol_filter <https://docs.scipy.org/doc/scipy-0.15.1/reference/generated/scipy.signal.savgol_filter.html>`_.

        Parameters
        ----------
        window_length : int
            Length of window that is used in the algorithm, must be positive odd integer.
        poly_order : int
            Polynomial order used in the algorithm, must be >= 1.
        """
        window_length = int(window_length)
        poly_order = int(poly_order)

        if poly_order < 1:
            raise ValueError("Polynomial order must be > 0.")

        if poly_order >= window_length:
            raise ValueError("Polynomial order must be less than window_length.")

        # window_length must be odd number
        window_length += int(not window_length % 2)

        self.data[:, 1] = savgol_filter(self.data[:, 1], window_length, poly_order)

        return self

    @add_modif_func(True, False)
    def gradient(self, edge_order=1):
        """
        Calculates a gradient using central differences and replaces the original values.

        Parameters
        ----------
        edge_order : int
            1 or 2, Gradient is calculated using N-th order accurate differences at the boundaries. Default: 1.
        """

        self.data[:, 1] = np.gradient(self.data[:, 1], self.data[:, 0], edge_order=edge_order)

        return self

    @add_modif_func(True, True)
    def differentiate(self, n=1):
        """
        Calculates a derivative spectrum and replaces the values. Length of the spectrum decreases by 1.

        Parameters
        ----------
        n : int
            The number of time s the values are differentiated. Default 1.
        """

        new_data = np.zeros((self.data.shape[0] - n, 2))

        x_diffs = np.diff(self.data[:, 0], 1)[n - 1:]  # changes in x data

        new_data[:, 1] = np.diff(self.data[:, 1], n) / (x_diffs ** n)  # dy/dx
        new_data[:, 0] = self.data[n:, 0]

        self.data = new_data

        return self

    @add_modif_func(True, False)
    def integrate(self, int_constant=0):
        """Integrates the spectrum using treapezoidal integration method, spectrum y values will
        be replaced by integrated values.

        Parameters
        ----------
        int_constant : {int, float}
            Integration constant, default 0.
        """
        self.data[:, 1] = cumtrapz(self.data[:, 1], self.data[:, 0], initial=int_constant)

        return self

    def _get_start_end_indexes(self, x0=None, x1=None):
        start = 0
        end = self.data.shape[0]

        if x0 is not None and x1 is not None and x0 > x1:
            x0, x1 = x1, x0

        if x0 is not None:
            start = fi(self.data[:, 0], x0)

        if x1 is not None:
            end = fi(self.data[:, 0], x1) + 1

        return start, end

    def _get_xy_from_range(self, x0=None, x1=None):
        start, end = self._get_start_end_indexes(x0, x1)

        x = self.data[start:end, 0]
        y = self.data[start:end, 1]

        return x, y

    @add_op_func()
    def integral(self, x0=None, x1=None, method: str = 'trapz'):
        """Calculate integral of the spectrum at specified x range [x0, x1] by integration method.
        If x0, x1 are None, all spectrum will be integrated.

        Parameters
        ----------
        x0 : {int, float, None}
            First x value. If both x0 and x1 are None, all spectrum will be integrated.
        x1 : {int, float, None}
            Last x value. If both x0 and x1 are None, all spectrum will be integrated.

        method : str
            this can be sum, trapz or simps
            default trapz

        Returns
        -------
        out : float
            Integral - scalar.
        """

        x, y = self._get_xy_from_range(x0, x1)

        if method == 'sum':
            x_diff = x[1:] - x[:-1]
            return (y[:-1] * x_diff).sum()
        elif method == 'trapz':
            return np.trapz(y, x)
        else:  # simpsons rule
            return simps(y, x)

    @add_modif_func(True, False)
    def baseline_correct(self, x0=None, x1=None):
        """Subtracts the average of y data from the specified x range [x0, x1] from y values.

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """

        x, y = self._get_xy_from_range(x0, x1)

        # subtract the average from y values
        self.data[:, 1] -= np.average(y)

        return self

    @add_modif_func(True, False)
    def normalize(self, x0=None, x1=None):
        """Finds an y maximum at specified x range [x0, x1] and divide all y values by this maximum.

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """

        x, y = self._get_xy_from_range(x0, x1)

        # normalize y values
        self.data[:, 1] /= y.max()

        return self

    @add_op_func()
    def find_maximum(self, x0=None, x1=None):
        """Returns a point (x and y value), which belongs to a maximum y value in a specified x range [x0, x1].

        Parameters
        ----------
        x0 : {int, float, None}
            First x value. If both x0 and x1 are None, all spectrum will be searched.
        x1 : {int, float, None}
            Last x value. If both x0 and x1 are None, all spectrum will be searched.

        Returns
        -------
        out : tuple - (x, y)
            A highest point in spectrum.
        """

        x, y = self._get_xy_from_range(x0, x1)

        max_idx = np.argmax(y)

        # return the x a y max value
        return x[max_idx], y[max_idx]

    @add_modif_func(True, True)
    def interpolate(self, spacing=1, kind='linear'):
        """Interpolates the spectrum, based on
        `scipy.interpolation.interp1d <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>`_.

        Parameters
        ----------
        spacing : {int, float}
            Sets the spacing between output x values, default 1.
        kind : str
            Sets the kind of interpolation method used, eg. 'linear', 'quadratic', etc. For more, refer to
            `interp1d documentation <https://docs.scipy.org/doc/scipy/reference/generated/scipy.interpolate.interp1d.html#scipy.interpolate.interp1d>`_.
            Default 'linear'.
        """

        data_min = self.data[0, 0]
        data_max = self.data[-1, 0]

        if spacing > data_max - data_min:
            raise ValueError(f"Spacing ({spacing}) cannot be larger that data itself.")

        # new x data must lie inside a range of original x values
        x_min = spacing * int(np.ceil(data_min / spacing))
        x_max = spacing * int(np.floor(data_max / spacing))

        # length of array of new data
        n = int((x_max - x_min) / spacing + 1)

        new_data = np.zeros((n, 2))
        new_data[:, 0] = np.linspace(x_min, x_max, num=n)  # new x values

        # interp1d returns a function that takes x values as the argument and returns the interpolated y values
        f = interp1d(self.data[:, 0], self.data[:, 1], kind=kind, copy=False, assume_sorted=True)

        new_data[:, 1] = f(new_data[:, 0])  # interpolate the y values

        # change the reference of self.data to new_data
        self.data = new_data

        return self

    @add_modif_func(True, True)
    def cut(self, x0=None, x1=None):
        """Cuts the spectrum to x range [x0, x1].

        Parameters
        ----------
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """

        if x0 is None and x1 is None:  # nothing to cut
            return self

        x, y = self._get_xy_from_range(x0, x1)

        self.data = np.vstack((x, y)).T

        return self

    @add_modif_func(True, False)
    def ceil_larger_to(self, value: float, x0=None, x1=None):
        """Values larger than value in the range [x0, x1] will be assigned to value .

        Parameters
        ----------
        value : {int, float}
            Value to ceil to.
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """

        start, end = self._get_start_end_indexes(x0, x1)

        y = self.data[start:end, 1]
        y[y > value] = value

        self.data[start:end, 1] = y

        return self

    @add_modif_func(True, False)
    def floor_lower_to(self, value: float, x0=None, x1=None):
        """Values lower than value in the range [x0, x1] will be assigned to a value .

        Parameters
        ----------
        value : {int, float}
            Value to floor to.
        x0 : {int, float}
            First x value.
        x1 : {int, float}
            Last x value.
        """

        start, end = self._get_start_end_indexes(x0, x1)

        y = self.data[start:end, 1]
        y[y < value] = value

        self.data[start:end, 1] = y

        return self

    @add_modif_func(True, True)
    def extend_by_value(self, x0=None, x1=None, value=0):
        """
        Extends this spectrum to a new x range [x0, x1] by zeros. The original data will
        be replaced. Spacing will be determined from current data.

        Parameters
        ----------
        x0 : {int, float}
            New fist x value.
        x1 : {int, float}
            New last x value.
        """

        if x0 is None and x1 is None:
            return self

        start_idx = 0
        end_idx = self.data.shape[0]

        if x0 is not None and x1 is not None and x0 > x1:
            x0, x1 = x1, x0

        if x0 is not None:
            start_idx = fi(self.data[:, 0], x0)

        if x1 is not None:
            end_idx = fi(self.data[:, 0], x1) + 1

        if start_idx != 0 and end_idx != self.data.shape[0]:
            # nothing to extend
            return self

        x_min = self.data[0, 0]
        x_max = self.data[-1, 0]

        x_dif = x_max - x_min
        spacing = x_dif / (self.data.shape[0] - 1)

        new_x_min = spacing * int(np.round(x0 / spacing, 0))

        min_stack = None
        try:
            num_min = int((x_min - new_x_min) / spacing + 1)
            min_lin_space = np.linspace(new_x_min, x_min, num=num_min)

            min_stack = np.ones((2, num_min - 1)) * value
            min_stack[0] = min_lin_space[:-1]
            min_stack = min_stack.T
        except ValueError:
            pass

        max_stack = None
        try:
            new_x_max = spacing * int(np.round(x1 / spacing, 0))

            num_max = int((new_x_max - x_max) / spacing + 1)
            max_lin_space = np.linspace(x_max, new_x_max, num=num_max)

            max_stack = np.ones((2, num_max - 1)) * value
            max_stack[0] = max_lin_space[1:]
            max_stack = max_stack.T
        except ValueError:
            pass

        if min_stack is not None and max_stack is not None:
            result = np.vstack((min_stack, self.data, max_stack))
        elif min_stack is not None:
            result = np.vstack((min_stack, self.data))
        else:
            result = np.vstack((self.data, max_stack))

        self.data = result

        return self

    def _operation_check(self, other):
        """Checks if the shapes of self and other data array are the same and where
        first and last x value is the same. If not, raises ValueError."""
        if self.data.shape[0] != other.data.shape[0] or not np.allclose(self.data[:, 0], other.data[:, 0]):
            raise ValueError(
                f"Spectra '{self.name}' and '{other.name}' does not share the same x values "
                f"(or have different dimension). Unable to perform operation.")

    def _arithmetic_operation(self, other, func_operation):
        if not isinstance(other, (Spectrum, int, float, list, tuple, np.ndarray)):
            return None, None

        # another spectrum
        if isinstance(other, Spectrum):
            self._operation_check(other)
            y_data = func_operation(self.data[:, 1], other.data[:, 1])
            other_str = other.name

        # list, tuple or ndarray that has the same dimension as current spectrum
        elif isinstance(other, (list, tuple, np.ndarray)):
            y = np.asarray(other)
            if y.shape[0] != self.data.shape[0]:
                raise ValueError(
                    f"Spectrum '{self.name}' and data does not have the same dimension. Unable to perform operation.")

            y_data = func_operation(self.data[:, 1], y)
            other_str = repr(other)

        else:  # number
            y_data = func_operation(self.data[:, 1], other)
            other_str = repr(other)

        max_len = 20
        if len(other_str) > max_len:
            other_str = other_str[:max_len]

        return Spectrum.from_xy_values(self.data[:, 0], np.nan_to_num(y_data)), other_str

    def __repr__(self):
        if self.data is None:
            return f'{self.__class__.__name__} without data'
        return f"{self.__class__.__name__}({self.name}, x_range=({self.data[0, 0]:.4g}, {self.data[-1, 0]:.4g}), " \
               f"n_points={self.data.shape[0]})"

    def __copy__(self):
        """Deep copy the current instance as Spectrum object."""
        return Spectrum(self.data.copy(), filepath=self.filepath, name=self.name, assume_sorted=True)


